from .model_base import JwstDataModel
from .image import ImageModel
from .slit import SlitModel, SlitDataModel


__all__ = ['MultiSlitModel']


class MultiSlitModel(JwstDataModel):
    """
    A data model for multi-slit images.

    This model has a special member `slits` that can be used to
    deal with an entire slit at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import SlitModel
       >>> multislit_model = MultiSlitModel()
       >>> multislit_model.slits.append(SlitModel())
       >>> multislit_model[0]
       <SlitModel>

    If ``init`` is a file name or an ``ImageModel`` or a ``SlitModel``instance,
    an empty ``SlitModel`` will be created and assigned to attribute ``slits[0]``,
    and the `data`, ``dq``, ``err``, ``var_rnoise``, and ``var_poisson``
    attributes from the input file or model will be copied to the
    first element of ``slits``.

    Parameters
    __________
    slits.items.data : numpy float32 array
         The science data

    slits.items.dq : numpy uint32 array
         Data quality array

    slits.items.err : numpy float32 array
         Error array

    slits.items.var_poisson : numpy float32 array
         variance due to poisson noise

    slits.items.var_rnoise : numpy float32 array
         variance due to read noise

    slits.items.wavelength : numpy float32 array
         Wavelength array, corrected for zero-point

    slits.items.barshadow : numpy float32 array
         Bar shadow correction

    slits.items.flatfield_point : numpy float32 array
         flatfield array for point source

    slits.items.flatfield_uniform : numpy float32 array
         flatfield array for uniform source

    slits.items.pathloss_point : numpy float32 array
         pathloss array for point source

    slits.items.pathloss_uniform : numpy float32 array
         pathloss array for uniform source

    slits.items.photom_point : numpy float32 array
         photom array for point source

    slits.items.photom_uniform : numpy float32 array
         photom array for uniform source

    slits.items.area : numpy float32 array
         Pixel area map array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/multislit.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, (SlitModel, ImageModel)):
            super(MultiSlitModel, self).__init__(init=None, **kwargs)
            self.update(init)
            slitdata = SlitDataModel(init)
            self.slits.append(slitdata)
            return

        super(MultiSlitModel, self).__init__(init=init, **kwargs)

    def __getitem__(self, key):
        """
        Returns a metadata value using a dotted name or
        a ``SlitModel``.
        """
        if isinstance(key, str) and key.split('.')[0] == 'meta':
            res = super(MultiSlitModel, self).__getitem__(key)
            return res
        elif isinstance(key, int):
            # Return an instance of a SlitModel
            # This only executes when the top object level
            # is called: object[1].key not object.slits[key]
            try:
                slit = self.slits[key]  # returns an ObjectNode instance
            except IndexError:
                raise ("Slit {0} doesn't exist".format(key))
            kwargs = {}
            items = dict(slit.items())
            for key in items:
                if not key.startswith(('meta', 'extra_fits')):
                    kwargs[key] = items[key]
            s = SlitModel(**kwargs)
            s.update(self)

            if slit.meta.hasattr('wcs'):
                s.meta.wcs = slit.meta.wcs
            return s
        else:
            raise ValueError("Invalid key {0}".format(key))



"""
This module supports the entry points for ASDF support for the `jwst.datamodels`.
"""

import sys

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


from asdf.resource import DirectoryResourceMapping
from stdatamodels.jwst import datamodels


def get_resource_mappings():
    """
    Get the `jwst.datamodels` resource mappings, that is the schemas for the datamodels.

    This method is registered with the `asdf.resource_mappings` entry point for
    the `jwst_datamodel`.

    Returns
    -------
    list of the `asdf.resource.ResourceMapping` instances containing the `jwst.datamodels`
    schemas.
    """
    resources_root = importlib_resources.files(datamodels)
    if not resources_root.is_dir():
        raise RuntimeError(f"Missing resources directory: {resources_root=}")

    return [
        DirectoryResourceMapping(
            resources_root / "schemas",
            "http://stsci.edu/schemas/jwst_datamodel/",
        ),
        DirectoryResourceMapping(
            resources_root / "metaschema",
            "http://stsci.edu/schemas/fits-schema/",
        )
    ]



from .reference import ReferenceFileModel


__all__ = ['DrizParsModel']


class DrizParsModel(ReferenceFileModel):
    """
    A data model for drizzle parameters reference tables.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/drizpars.schema"



from .reference import ReferenceFileModel

__all__ = ['MirMrsPtCorrModel']


class MirMrsPtCorrModel(ReferenceFileModel):
    """
    A data model for MIRI mrs IFU across-slice corrections file.

    Parameters
    __________
    init : any
        Any of the initializers supported by `~jwst.datamodels.DataModel`.

    data : numpy array
        An array-like object containing the pixel-by-pixel spectral leak values
        in units of (MJy / sr) / (DN / sec).

    err : numpy array
        An array-like object containing the uncertainties in the spectral leak
        values, in the same units as the data array.

    dq : numpy array
        An array-like object containing bit-encoded data quality flags,
        indicating problem conditions for values in the data array.

    dq_def : numpy array
        A table-like object containing the data quality definitions table.

    tracor_table : numpy table
         IFU across slice transmission correction

    wavcorr_optical_table : numpy table
         IFU across slice wavelength offset table 1

    wavcorr_xslice_table : numpy table
         IFU across slice wavelength offset table 2

    wavcorr_shift_table : numpy table
         IFU across slice wavelength offset table 3

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/miri_mrsptcorr.schema"



from .model_base import JwstDataModel
from .spec import SpecModel


__all__ = ['MultiSpecModel']


class MultiSpecModel(JwstDataModel):
    """
    A data model for multi-spec images.

    This model has a special member `spec` that can be used to
    deal with an entire spectrum at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import SpecModel
       >>> multispec_model = MultiSpecModel()
       >>> multispec_model.spec.append(SpecModel())
       >>> multispec_model.spec[0] # doctest: +SKIP
       <SpecModel>

    If `init` is a `SpecModel` instance, an empty `SpecModel` will be
    created and assigned to attribute `spec[0]`, and the `spec_table`
    attribute from the input `SpecModel` instance will be copied to
    the first element of `spec`.  `SpecModel` objects can be appended
    to the `spec` attribute by using its `append` method.

    Parameters
    __________
    int_times : numpy table
         table of times for each integration

    spec.items.spec_table : numpy table
         Extracted spectral data table

    Examples
    --------
    >>> output_model = MultiSpecModel()
    >>> spec = SpecModel()       # for the default data type
    >>> for slit in input_model.slits:  # doctest: +SKIP
    ...     slitname = slit.name
    ...     slitmodel = ExtractModel()
    ...     slitmodel.fromJSONFile(extref, slitname)
    ...     column, wavelength, countrate = slitmodel.extract(slit.data)
    ...     otab = np.array(zip(column, wavelength, countrate),
    ...                     dtype=spec.spec_table.dtype)
    ...     spec = datamodels.SpecModel(spec_table=otab)
    ...     output_model.spec.append(spec)
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/multispec.schema"

    def __init__(self, init=None, **kwargs):

        if isinstance(init, SpecModel):
            super(MultiSpecModel, self).__init__(init=None, **kwargs)
            self.spec.append(self.spec.item())
            self.spec[0].spec_table = init.spec_table
            return

        super(MultiSpecModel, self).__init__(init=init, **kwargs)



import pytest
from astropy.io import fits
import numpy as np
from numpy.testing import assert_array_almost_equal
import yaml

from stdatamodels import asdf_in_fits

from models import FitsModel


def test_write_linked(tmp_path):
    hdulist = fits.HDUList()
    sci = np.arange(512, dtype=float)
    dq = np.arange(512, dtype=float) + 1
    hdulist.append(fits.ImageHDU(sci, name='SCI'))
    hdulist.append(fits.ImageHDU(dq, name='DQ'))

    tree = {
        'meta': {
            'foo': 'bar',
        },
        'model': {
            'sci': {
                'data': hdulist['SCI'].data,
            },
            'dq': {
                'data': hdulist['DQ'].data,
            }
        }
    }

    fn = tmp_path / "test.fits"

    asdf_in_fits.write(fn, tree, hdulist)

    # confirm it matches
    with asdf_in_fits.open(fn) as af:
        assert_array_almost_equal(af['model']['sci']['data'], sci)
        assert_array_almost_equal(af['model']['dq']['data'], dq)
        assert af['meta']['foo'] == 'bar'

    with fits.open(fn) as read_hdulist:
        # hdu should have sci, dq, and ASDF
        assert len(read_hdulist) == 3
        # check asdf extension has no blocks by looking for bytes
        # after the end of the yaml document for the tree
        bs = read_hdulist['ASDF'].data['ASDF_METADATA'].tobytes()
        assert len(bs.strip().split(b'...')[1]) == 0


def test_write_asdf_in_fits_no_hdulist(tmp_path):
    sci = np.arange(512, dtype=float)
    dq = np.arange(512, dtype=float) + 1
    tree = {
        'meta': {
            'foo': 'bar',
        },
        'model': {
            'sci': {
                'data': sci,
            },
            'dq': {
                'data': dq,
            }
        }
    }
    fn = tmp_path / "test.fits"

    asdf_in_fits.write(fn, tree)

    # confirm it matches
    with asdf_in_fits.open(fn) as af:
        assert_array_almost_equal(af['model']['sci']['data'], sci)
        assert_array_almost_equal(af['model']['dq']['data'], dq)
        assert af['meta']['foo'] == 'bar'

    with fits.open(fn) as read_hdulist:
        # hdu should have primary and ASDF
        assert len(read_hdulist) == 2
        # check asdf extension has blocks by looking at the block index
        bs = read_hdulist['ASDF'].data['ASDF_METADATA'].tobytes()
        block_index_bs = bs.split(b'BLOCK INDEX')[1].strip()
        block_offsets = yaml.load(block_index_bs, yaml.SafeLoader)
        assert len(block_offsets) == 2


def test_write_asdf_in_fits_partial_hdulist(tmp_path):
    hdulist = fits.HDUList()
    # add other data to hdulist
    hdulist.append(fits.PrimaryHDU())

    sci = np.arange(512, dtype=float)
    dq = np.arange(512, dtype=float) + 1
    tree = {
        'meta': {
            'foo': 'bar',
        },
        'model': {
            'sci': {
                'data': sci,
            },
            'dq': {
                'data': dq,
            }
        }
    }
    fn = tmp_path / "test.fits"

    asdf_in_fits.write(fn, tree, hdulist)

    # confirm it matches
    with asdf_in_fits.open(fn) as af:
        assert_array_almost_equal(af['model']['sci']['data'], sci)
        assert_array_almost_equal(af['model']['dq']['data'], dq)
        assert af['meta']['foo'] == 'bar'

    with fits.open(fn) as read_hdulist:
        # hdu should have primary and ASDF
        assert len(read_hdulist) == 2
        # check asdf extension has blocks by looking at the block index
        bs = read_hdulist['ASDF'].data['ASDF_METADATA'].tobytes()
        block_index_bs = bs.split(b'BLOCK INDEX')[1].strip()
        block_offsets = yaml.load(block_index_bs, yaml.SafeLoader)
        assert len(block_offsets) == 2


@pytest.fixture
def test_array():
    return np.arange(1000, dtype='f4').reshape((100, 10))


@pytest.fixture
def saved_array_model(tmp_path, test_array):
    # make a model
    fn = tmp_path / "test.fits"
    m = FitsModel(test_array)
    m.save(fn)
    return fn


def test_open_asdf_in_fits(saved_array_model, test_array):
    # read it back (not in a context)
    af = asdf_in_fits.open(saved_array_model)
    # confirm it matches
    assert_array_almost_equal(af['data'], test_array)
    af.close()


def test_open_asdf_in_fits_context(saved_array_model, test_array):
    # read it back as a context
    with asdf_in_fits.open(saved_array_model) as af:
        # confirm it matches
        assert_array_almost_equal(af['data'], test_array)


def test_open_asdf_in_fits_hdu(saved_array_model, test_array):
    """
    Test that asdf_in_fits.open can read from an already
    opened HDUList and that it does not close the HDUList
    when the AsdfFile is closed
    """
    with fits.open(saved_array_model) as hdu:
        with asdf_in_fits.open(hdu) as af:
            assert_array_almost_equal(af['data'], test_array)
        # make sure file was not closed with context
        assert not hdu.fileinfo(0)['file'].closed



from .reference import ReferenceFileModel


__all__ = ['RSCDModel']


class RSCDModel(ReferenceFileModel):
    """
    A data model for the RSCD reference file.

    Parameters
    __________
    rscd_group_skip_table : numpy table
        Reference table for RSCD correction baseline correction
        A table with 3 columns that set the number of groups to skip for each
        subarray and readpatt

    rscd_gen_table : numpy table
        Reference table for RSCD correction enhanced correction
        A table with 5 columns that sets up general parameters for the enhanced
        correction

    rscd_int1_table : numpy table
        Reference table for RSCD correction enhanced correction
        A table with 7 columns that sets up correction parameters for the enhanced
        correction for integration 1 based on subarray, readpatt, even or odd row

    rscd_int2_table : numpy table
        Reference table for RSCD correction enhanced correction
        A table with 7 columns that sets up correction parameters for the enhanced
        correction for integration 2 based on subarray, readpatt, even or odd row

    rscd_int3_table : numpy table
        Reference table for RSCD correction enhanced correction
        A table with 7 columns that sets up correction parameters for the enhanced
        correction for integration 3 based on subarray, readpatt, even or odd row

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/rscd.schema"



from .reference import ReferenceFileModel

__all__ = ['IPCModel']


class IPCModel(ReferenceFileModel):
    """
    A data model for IPC kernel checking information.

    Parameters
    __________
    data : numpy float32 array
         IPC deconvolution kernel
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ipc.schema"



import os
import logging
from collections.abc import Iterable
from stdatamodels.jwst import datamodels as dm
import pytest
import warnings

os.environ["CRDS_SERVER_URL"] = "https://jwst-crds.stsci.edu"

# the CRDS_SERVER_URL environment variable must be set before
# crds is imported
import crds  # noqa: E402
from crds.client.api import cache_references, dump_files  # noqa: E402
from crds.core.exceptions import IrrelevantReferenceTypeError  # noqa: E402


log = logging.getLogger(__name__)


def flatten(xs):
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


ignored_parkeys = ['META.OBSERVATION.DATE',
                   'META.OBSERVATION.TIME',
                   ]

apcorr_model_map = {
    'MIR_LRS-FIXEDSLIT': dm.MirLrsApcorrModel,
    'MIR_LRS-SLITLESS': dm.MirLrsApcorrModel,
    'MIR_MRS': dm.MirMrsApcorrModel,
    'MIR_IMAGE': dm.MirImgApcorrModel,
    'NRC_GRISM': dm.NrcWfssApcorrModel,
    'NRC_WFSS': dm.NrcWfssApcorrModel,
    'NRC_IMAGE': dm.NrcImgApcorrModel,
    'NIS_WFSS': dm.NisWfssApcorrModel,
    'NIS_IMAGE': dm.NisImgApcorrModel,
    'NRS_BRIGHTOBJ': dm.NrsFsApcorrModel,
    'NRS_FIXEDSLIT': dm.NrsFsApcorrModel,
    'NRS_IFU': dm.NrsIfuApcorrModel,
    'NRS_MSASPEC': dm.NrsMosApcorrModel,
    'FGS_IMAGE': dm.FgsImgApcorrModel,
    'FGS': dm.FgsImgApcorrModel,
    'NIRCAM': dm.NrcImgApcorrModel,
    'NIRISS': dm.NisWfssApcorrModel,
}

area_model_map = {
    'NRS_MSASPEC': dm.NirspecMosAreaModel,
    'NRS_FIXEDSLIT': dm.NirspecSlitAreaModel,
    'NRS_IFU': dm.NirspecIfuAreaModel,
    'other': dm.PixelAreaModel,
}

cubepar_model_map = {
    'NIRSPEC': dm.NirspecIFUCubeParsModel,
    'MIRI': dm.MiriIFUCubeParsModel,
}

distortion_model_map = {
    'MIR_MRS': dm.DistortionMRSModel,
    'other': dm.DistortionModel,
}

flat_model_map = {
    'NIRSPEC': dm.NirspecFlatModel,
    'other': dm.FlatModel,
}

pathloss_model_map = {
    'MIR_LRS': dm.MirLrsPathlossModel,
    'other': dm.PathlossModel,
}

photom_model_map = {
    'MIR_LRS-FIXEDSLIT': dm.MirLrsPhotomModel,
    'MIR_LRS-SLITLESS': dm.MirLrsPhotomModel,
    'MIR_MRS': dm.MirMrsPhotomModel,
    'MIR_IMAGE': dm.MirImgPhotomModel,
    'NRC_GRISM': dm.NrcWfssPhotomModel,
    'NRC_WFSS': dm.NrcWfssPhotomModel,
    'NRC_IMAGE': dm.NrcImgPhotomModel,
    'NIS_WFSS': dm.NisWfssPhotomModel,
    'NIS_SOSS': dm.NisSossPhotomModel,
    'NIS_IMAGE': dm.NisImgPhotomModel,
    'NRS_BRIGHTOBJ': dm.NrsFsPhotomModel,
    'NRS_FIXEDSLIT': dm.NrsFsPhotomModel,
    'NRS_IFU': dm.NrsMosPhotomModel,
    'NRS_MSASPEC': dm.NrsMosPhotomModel,
    'FGS_IMAGE': dm.FgsImgPhotomModel,
    'FGS': dm.FgsImgPhotomModel,
    'NIRCAM': dm.NrcImgPhotomModel,
    'MIRI': dm.MirImgPhotomModel,
    'NIRISS': dm.NisWfssPhotomModel,
}

resol_model_map = {
    'MIRI': dm.MiriResolutionModel,
    'other': dm.ResolutionModel,
}

ref_to_multiples_dict = {
    'apcorr': apcorr_model_map,
    'area': area_model_map,
    'cubepar': cubepar_model_map,
    'distortion': distortion_model_map,
    'flat': flat_model_map,
    'pathloss': pathloss_model_map,
    'photom': photom_model_map,
    'resol': resol_model_map,
}

ref_to_datamodel_dict = {
    'abvegaoffset': dm.ABVegaOffsetModel,
    'barshadow': dm.BarshadowModel,
    'camera': dm.CameraModel,
    'collimator': dm.CollimatorModel,
    'dark': dm.DarkModel,
    'dflat': dm.NirspecFlatModel,
    'disperser': dm.DisperserModel,
    'drizpars': dm.DrizParsModel,
    'extract1d': dm.Extract1dIFUModel,
    'fflat': dm.NirspecFlatModel,
    'filteroffset': dm.FilteroffsetModel,
    'fore': dm.FOREModel,
    'fpa': dm.FPAModel,
    'fringe': dm.FringeModel,
    'fringefreq': dm.FringeFreqModel,
    'gain': dm.GainModel,
    'ifufore': dm.IFUFOREModel,
    'ifupost': dm.IFUPostModel,
    'ifuslicer': dm.IFUSlicerModel,
    'ipc': dm.IPCModel,
    'lastframe': dm.LastFrameModel,
    'linearity': dm.LinearityModel,
    'mask': dm.MaskModel,
    'mrsxartcorr': dm.MirMrsXArtCorrModel,
    'mrsptcorr': dm.MirMrsPtCorrModel,
    'msa': dm.MSAModel,
    'msaoper': None,
    'ote': dm.OTEModel,
    'persat': dm.PersistenceSatModel,
    'psfmask': dm.PsfMaskModel,
    'readnoise': dm.ReadnoiseModel,
    'refpix': dm.IRS2Model,
    'regions': dm.RegionsModel,
    'reset': dm.ResetModel,
    'rscd': dm.RSCDModel,
    'saturation': dm.SaturationModel,
    'sflat': dm.NirspecFlatModel,
    'speckernel': dm.SpecKernelModel,
    'specprofile': dm.SpecProfileModel,
    'spectrace': dm.SpecTraceModel,
    'specwcs': dm.SpecwcsModel,
    'straymask': dm.StrayLightModel,
    'superbias': dm.SuperBiasModel,
    'throughput': dm.ThroughputModel,
    'trapdensity': dm.TrapDensityModel,
    'trappars': dm.TrapParsModel,
    'tsophot': dm.TsoPhotModel,
    'wavecorr': dm.WaveCorrModel,
    'wavelengthrange': dm.WavelengthrangeModel,
    'wavemap': dm.WaveMapModel,
    'wcsregions': None,
    'wfssbkg': dm.WfssBkgModel,
}


@pytest.mark.skipif(
    "config.getoption('--no-crds')",
    reason="no_crds option was enabled",
)
@pytest.mark.parametrize('instrument', ['fgs', 'miri', 'nircam', 'niriss', 'nirspec'])
def test_crds_selectors_vs_datamodel(jail_environ, instrument):
    log.info(f"crds_path: {crds.config.get_crds_path()}")
    log.info(f"crds_server: {crds.config.get_server_url('jwst')}")

    context = crds.get_context_name('jwst')
    dump_files(context)
    pmap = crds.get_cached_mapping(context)

    imap = pmap.get_imap(instrument)
    log.info(f"Beginning tests for {instrument}")

    # get the reftypes
    reftypes = imap.get_filekinds()
    # remove pars- files
    _ = [reftypes.remove(name) for name in reftypes[::-1] if name.startswith('pars-')]

    # iterate over reftypes for this instrument
    for reftype in reftypes:
        try:
            r = imap.get_rmap(reftype)
            parkeys = [p for p in list(flatten(list(r.parkey))) if p not in ignored_parkeys]
            log.debug(f"Parkeys for {reftype}: {parkeys}")
            for f in r.reference_names():
                # Ensure filetype is kind to be loaded into datamodel
                if 'fits' in f or 'asdf' in f:
                    # Find datamodel appropriate for this reference file
                    # If reftype has multiple datamodels possible, do some guesswork
                    if reftype in ref_to_multiples_dict.keys():
                        model_map = ref_to_multiples_dict[reftype]
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore', dm.util.NoTypeWarning)
                            refs = cache_references(context, {reftype: f})
                            with dm.open(refs[reftype]) as model:
                                try:
                                    ref_exptype = model.meta.exposure.type
                                except AttributeError:
                                    ref_exptype = None
                                ref_instrument = model.meta.instrument.name
                        if ref_exptype in model_map.keys():
                            ref_model = model_map[ref_exptype]
                        elif ref_instrument in model_map.keys():
                            ref_model = model_map[ref_instrument]
                        else:
                            ref_model = model_map['other']
                    # Simple one to one translation of reftype to datamodel
                    else:
                        ref_model = ref_to_datamodel_dict[reftype]

                    log.debug(f"Loading {reftype} reference for {instrument} as {ref_model}")

                    if ref_model is None:
                        log.warning(f"No datamodel found for {reftype}: skipping...")
                        break
                    # No need to actually load the reference file into the datamodel!
                    with ref_model() as m:
                        for key in parkeys:
                            assert len(m.search_schema(key.lower())) > 0
                    break
        except IrrelevantReferenceTypeError as e:
            log.debug(e)
            pass



# from .jwst_models import (Gwa2SlitConverter, Slit2MsaConverter, LogicalConverter,
#                           NirissSOSSConverter, RefractionIndexConverter,
#                           MIRI_AB2SliceConverter, NIRCAMGrismDispersionConverter,
#                           NIRISSGrismDispersionConverter, GratingEquationConverter,
#                           V23ToSkyConverter, SnellConverter, CoordsConverter,
#                           Rotation3DToGWAConverter)
#
# __all__ = ['CoordsConverter', 'GratingEquationConverter', 'Gwa2SlitConverter',
#            'LogicalConverter', 'MIRI_AB2SliceConverter', 'NIRCAMGrismDispersionConverter',
#            'NIRISSGrismDispersionConverter', 'NirissSOSSConverter', 'RefractionIndexConverter',
#            'Rotation3DToGWAConverter', 'Slit2MsaConverter', 'SnellConverter',
#            'V23ToSkyConverter']



import warnings

from stdatamodels.validate import ValidationWarning
from stdatamodels.dynamicdq import dynamic_mask

from .model_base import JwstDataModel
from .dqflags import pixel


__all__ = ['ReferenceFileModel', "ReferenceImageModel", "ReferenceCubeModel",
           "ReferenceQuadModel"]


class ReferenceFileModel(JwstDataModel):
    """
    A data model for reference tables

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/referencefile.schema"

    def __init__(self, init=None, **kwargs):
        super(ReferenceFileModel, self).__init__(init=init, **kwargs)
        self._no_asdf_extension = True
        self.meta.telescope = "JWST"

    def validate(self):
        """
        Convenience function to be run when files are created.
        Checks that required reference file keywords are set.
        """
        to_fix = []
        to_check = ['description', 'reftype', 'author', 'pedigree', 'useafter']
        for field in to_check:
            if getattr(self.meta, field) is None:
                to_fix.append(field)
        if self.meta.instrument.name is None:
            to_fix.append('instrument.name')
        if self.meta.telescope != 'JWST':
            to_fix.append('telescope')
        if to_fix:
            self.print_err(f'Model.meta is missing values for {to_fix}')
        super().validate()

    def save(self, path, dir_path=None, *args, **kwargs):
        """
        Save data model.  If the 'dq' and 'dq_def' exist they need special
        handling.
        """
        if (self.hasattr('dq_def') and self.hasattr("dq")
           and self.dq is not None and self.dq_def is not None):
            # Save off uncompressed DQ array.  Compress DQ array
            # according to 'dq_def' for save.  Then restore
            # uncompressed DQ array.
            dq_orig = self.dq.copy()
            self.dq = dynamic_mask(self, pixel, inv=True)
            output_path = super().save(path, dir_path, *args, **kwargs)
            self.dq = dq_orig
        else:
            output_path = super().save(path, dir_path, *args, **kwargs)
        return output_path

    def print_err(self, message):
        if self._strict_validation:
            raise ValueError(message)
        else:
            warnings.warn(message, ValidationWarning)


class ReferenceImageModel(ReferenceFileModel):
    """
    A data model for 2D reference images.

    Reference image data model.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/referenceimage.schema"

    def __init__(self, init=None, **kwargs):
        super(ReferenceImageModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err

        if self.hasattr('dq_def'):
            self.dq = dynamic_mask(self, pixel)


class ReferenceCubeModel(ReferenceFileModel):
    """
    A data model for 3D reference images

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/referencecube.schema"

    def __init__(self, init=None, **kwargs):
        super(ReferenceCubeModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err


class ReferenceQuadModel(ReferenceFileModel):
    """
    A data model for 4D reference images

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/referencequad.schema"

    def __init__(self, init=None, **kwargs):
        super(ReferenceQuadModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



import os


def check(init):
    """
    Determine the type of a file and return it as a string

    Parameters
    ----------

    init : file path or file object

    Returns
    -------
    file_type: a string with the file type ("asdf", "asn", or "fits")

    """
    supported = ('asdf', 'fits', 'json')

    if isinstance(init, str):
        path, ext = os.path.splitext(init)
        ext = ext.strip('.')

        if not ext:
            raise ValueError(f'Input file path does not have an extension: {init}')

        if ext not in supported:  # Could be the file is zipped; try splitting again
            path, ext = os.path.splitext(path)
            ext = ext.strip('.')

            if ext not in supported:
                raise ValueError(f'Unrecognized file type for: {init}')

        if ext == 'json':  # Assume json input is an association
            return 'asn'

        return ext

    if hasattr(init, "read") and hasattr(init, "seek"):
        magic = init.read(5)
        init.seek(0, 0)

        if not magic or len(magic) < 5:
            raise ValueError(f"Cannot get file type of {str(init)}")

        if magic == b'#ASDF':
            return "asdf"

        if magic == b'SIMPL':
            return "fits"

        return "asn"

    raise ValueError(f"Cannot get file type of {str(init)}")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Models used by the JWST pipeline.

The models are written using the astropy.modeling framework.
Since they are specific to JWST, the models and their ASDF schemas
are kept here separately from astropy. An ASDF extension for this package is
registered with ASDF through entry points.
"""


import math
from collections import namedtuple
import numpy as np
from astropy.modeling.core import Model
from astropy.modeling.parameters import Parameter, InputParameterError
from astropy.modeling.models import (Rotation2D, Mapping, Tabular1D, Const1D)
from astropy.modeling.models import math as astmath
from astropy.utils import isiterable
from ...properties import ListNode


__all__ = ['Gwa2Slit', 'Slit2Msa', 'Logical', 'NirissSOSSModel', 'Slit',
           'NIRCAMForwardRowGrismDispersion', 'NIRCAMForwardColumnGrismDispersion',
           'NIRCAMBackwardGrismDispersion', 'MIRI_AB2Slice', 'GrismObject',
           'NIRISSForwardRowGrismDispersion', 'NIRISSForwardColumnGrismDispersion',
           'NIRISSBackwardGrismDispersion', 'V2V3ToIdeal', 'IdealToV2V3',
           'RefractionIndexFromPrism', 'Snell', 'V23ToSky', 'Rotation3DToGWA',
           'AngleFromGratingEquation', 'WavelengthFromGratingEquation']


N_SHUTTERS_QUADRANT = 62415
""" Number of shutters per quadrant in the NIRSPEC MSA shutter array"""


Slit = namedtuple('Slit', ["name", "shutter_id", "dither_position", "xcen", "ycen",
                           "ymin", "ymax", "quadrant", "source_id", "shutter_state",
                           "source_name", "source_alias", "stellarity",
                           "source_xpos", "source_ypos", "source_ra", "source_dec"])
""" Nirspec Slit structure definition"""


Slit.__new__.__defaults__ = ("", 0, 0, 0.0, 0.0, 0.0, 0.0, 0, 0, "", "", "",
                             0.0, 0.0, 0.0, 0.0, 0.0)


class GrismObject(namedtuple('GrismObject', ("sid",
                                             "order_bounding",
                                             "sky_centroid",
                                             "partial_order",
                                             "waverange",
                                             "sky_bbox_ll",
                                             "sky_bbox_lr",
                                             "sky_bbox_ur",
                                             "sky_bbox_ul",
                                             "xcentroid",
                                             "ycentroid",
                                             "is_extended",
                                             "isophotal_abmag",
                                             ), rename=False)):
    """ Grism Objects identified from a direct image catalog and segment map.

    Parameters
    ----------
    sid : int
        source identified
    xcentroid : float
        x center of object in pixels
    ycentroid : float
        y center of object in pixels
    order_bounding : dict{order: tuple}
        Contains the object x,y bounding locations on the image
        keyed on spectral order
    partial_order : bool
        True if the order is only partially contained on the image
    waverange : list
        wavelength range for the order
    is_extended : bool
        True if marked as extended in the source catalog
    isophotal_abmag : float
        isophotal_abmag from the source catalog
    sky_centroid: `~astropy.coordinates.SkyCoord`
        ra and dec of the center of the object
    sky_bbox_ll : `~astropy.coordinates.SkyCoord`
        Lower left corner of the minimum bounding box
    sky_bbox_lr : `~astropy.coordinates.SkyCoord`
        Lower right corder of the minimum bounding box
    sky_bbox_ul : `~astropy.coordinates.SkyCoord`
        Upper left corner of the minimum bounding box
    sky_bbox_ur : `~astropy.coordinates.SkyCoord`
        Upper right corner of the minimum bounding box

    Notes
    -----
    The object bounding box is computed from the segmentation map,
    using the min and max wavelength for each of the orders that
    are available. The order_bounding member is a dictionary of
    bounding boxes for the object keyed by order

    ra and dec are the sky ra and dec of the center of the object as measured
    from the non-dispersed image.

    order_bounding is stored as a lookup dictionary per order and contains
    the object x,y bounding location on the grism image
    GrismObject(order_bounding={"+1":((xmin,xmax),(ymin,ymax)),"+2":((2,3),(2,3))})

    """
    __slots__ = ()  # prevent instance dictionary for lower memory

    def __new__(cls,
                sid=None,
                order_bounding={},
                sky_centroid=None,
                partial_order={},
                waverange=None,
                sky_bbox_ll=None,
                sky_bbox_lr=None,
                sky_bbox_ur=None,
                sky_bbox_ul=None,
                xcentroid=None,
                ycentroid=None,
                is_extended=None,
                isophotal_abmag=None):

        return super(GrismObject, cls).__new__(cls,
                                               sid=sid,
                                               order_bounding=order_bounding,
                                               sky_centroid=sky_centroid,
                                               partial_order=partial_order,
                                               waverange=waverange,
                                               sky_bbox_ll=sky_bbox_ll,
                                               sky_bbox_lr=sky_bbox_lr,
                                               sky_bbox_ur=sky_bbox_ur,
                                               sky_bbox_ul=sky_bbox_ul,
                                               xcentroid=xcentroid,
                                               ycentroid=ycentroid,
                                               is_extended=is_extended,
                                               isophotal_abmag=isophotal_abmag)

    def __str__(self):
        """Return a pretty print for the object information."""
        return ("id: {0}\n"
                "order_bounding {1}\n"
                "sky_centroid: {2}\n"
                "sky_bbox_ll: {3}\n"
                "sky_bbox_lr: {4}\n"
                "sky_bbox_ur: {5}\n"
                "sky_bbox_ul:{6}\n"
                "xcentroid: {7}\n"
                "ycentroid: {8}\n"
                "partial_order: {9}\n"
                "waverange: {10}\n"
                "is_extended: {11}\n"
                "isophotal_abmag: {12}\n"
                .format(self.sid,
                        str(self.order_bounding),
                        str(self.sky_centroid),
                        str(self.sky_bbox_ll),
                        str(self.sky_bbox_lr),
                        str(self.sky_bbox_ur),
                        str(self.sky_bbox_ul),
                        self.xcentroid,
                        self.ycentroid,
                        str(self.partial_order),
                        str(self.waverange),
                        str(self.is_extended),
                        str(self.isophotal_abmag)))


class MIRI_AB2Slice(Model):
    """
    MIRI MRS alpha, beta to slice transform

    Parameters
    ----------
    beta_zero : float
    beta_del : float
    """
    standard_broadcasting = False
    _separable = False
    fittable = False

    n_inputs = 1
    n_outputs = 1

    beta_zero = Parameter('beta_zero', default=0)
    """ Beta_zero parameter"""
    beta_del = Parameter('beta_del', default=1)
    """ Beta_del parameter"""
    channel = Parameter("channel", default=1)
    """ MIRI MRS channel: one of 1, 2, 3, 4"""

    def __init__(self, beta_zero=beta_zero, beta_del=beta_del, channel=channel, **kwargs):
        super().__init__(beta_zero=beta_zero, beta_del=beta_del,
                         channel=channel, **kwargs)
        self.inputs = ("beta",)
        """ "beta": the beta angle """
        self.outputs = ("slice",)
        """ "slice": Slice number"""

    @staticmethod
    def evaluate(beta, beta_zero, beta_del, channel):
        s = channel * 100 + (beta - beta_zero) / beta_del + 1
        return _toindex(s)


class RefractionIndexFromPrism(Model):
    """
    Compute the refraction index of a prism (NIRSpec).

    Parameters
    ----------
    prism_angle : float
        Prism angle in deg.

    """
    standard_broadcasting = False
    _separable = False

    n_inputs = 3
    n_outputs = 1

    prism_angle = Parameter(setter=np.deg2rad, getter=np.rad2deg)

    def __init__(self, prism_angle, name=None):
        super(RefractionIndexFromPrism, self).__init__(prism_angle=prism_angle, name=name)
        self.inputs = ("alpha_in", "beta_in", "alpha_out",)
        self.outputs = ("n",)

    def evaluate(self, alpha_in, beta_in, alpha_out, prism_angle):
        sangle = (math.sin(prism_angle))
        cangle = (math.cos(prism_angle))
        nsq = ((alpha_out + alpha_in * (1 - 2 * sangle ** 2)) / (2 * sangle * cangle)) ** 2 + \
            alpha_in ** 2 + beta_in ** 2
        return np.sqrt(nsq)


class Gwa2Slit(Model):
    """
    NIRSpec GWA to slit transform.

    Parameters
    ----------
    slits : list
        A list of open slits.
        A slit is a namedtuple of type `~stdatamodels.jwst.transforms.models.Slit`
        Slit("name", "shutter_id", "dither_position", "xcen", "ycen", "ymin", "ymax",
        "quadrant", "source_id", "shutter_state", "source_name",
        "source_alias", "stellarity", "source_xpos", "source_ypos",
        "source_ra", "source_dec"])
    models : list
        List of models (`~astropy.modeling.core.Model`) corresponding to the
        list of slits.
    """
    _separable = False

    n_inputs = 4
    n_outputs = 4

    def __init__(self, slits, models):
        if isiterable(slits[0]):
            self._slits = [tuple(s) for s in slits]
            self.slit_ids = [s[0] for s in self._slits]
        else:
            self._slits = list(slits)
            self.slit_ids = self._slits

        self.models = models
        super(Gwa2Slit, self).__init__()
        self.inputs = ('name', 'angle1', 'angle2', 'angle3')
        """ Name of the slit and the three angle coordinates at the GWA going from detector to sky."""
        self.outputs = ('name', 'x_slit', 'y_slit', 'lam')
        """ Name of the slit, x and y coordinates within the virtual slit and wavelength."""

    @property
    def slits(self):
        if isiterable(self._slits[0]):
            return [Slit(*row) for row in self._slits]
        else:
            return self.slit_ids

    def get_model(self, name):
        index = self.slit_ids.index(name)
        return self.models[index]

    def evaluate(self, name, x, y, z):
        index = self.slit_ids.index(name)
        return (name, ) + self.models[index](x, y, z)


class Slit2Msa(Model):
    """
    Transform from Nirspec ``slit_frame`` to ``msa_frame``.

    Parameters
    ----------
    slits : list
        A list of open slits.
        A slit is a namedtuple, `~stdatamodels.jwst.transforms.models.Slit`
        Slit("name", "shutter_id", "dither_position", "xcen", "ycen", "ymin", "ymax",
        "quadrant", "source_id", "shutter_state", "source_name",
        "source_alias", "stellarity", "source_xpos", "source_ypos",
        "source_ra", "source_dec")
    models : list
        List of models (`~astropy.modeling.core.Model`) corresponding to the
        list of slits.
    """
    _separable = False

    n_inputs = 3
    n_outputs = 2

    def __init__(self, slits, models):
        super(Slit2Msa, self).__init__()
        self.inputs = ('name', 'x_slit', 'y_slit')
        """ Name of the slit, x and y coordinates within the virtual slit."""
        self.outputs = ('x_msa', 'y_msa')
        """ x and y coordinates in the MSA frame."""
        if isiterable(slits[0]):
            self._slits = [tuple(s) for s in slits]
            self.slit_ids = [s[0] for s in self._slits]
        else:
            self._slits = list(slits)
            self.slit_ids = self._slits
        self.models = models

    @property
    def slits(self):
        if isiterable(self._slits[0]):
            return [Slit(*row) for row in self._slits]
        else:
            return self.slit_ids

    def get_model(self, name):
        index = self.slit_ids.index(name)
        return self.models[index]

    def evaluate(self, name, x, y):
        index = self.slit_ids.index(name)
        return self.models[index](x, y)


class NirissSOSSModel(Model):
    """
    NIRISS SOSS wavelength solution model.

    Parameters
    ----------
    spectral_orders : list of int
        Spectral orders for which there is a wavelength solution.
    models : list of `~astropy.modeling.core.Model`
        A list of transforms representing the wavelength solution for
        each order in spectral orders. It should match the order in
        ``spectral_orders``.
    """

    _separable = False

    n_inputs = 3
    n_outputs = 3

    def __init__(self, spectral_orders, models):
        super(NirissSOSSModel, self).__init__()
        self.inputs = ('x', 'y', 'spectral_order')
        """ x and y pixel coordinates and spectral order"""
        self.outputs = ('ra', 'dec', 'lam')
        """ RA and DEC coordinates and wavelength"""

        self.spectral_orders = spectral_orders
        self.models = dict(zip(spectral_orders, models))

    def get_model(self, spectral_order):
        return self.models[spectral_order]

    def evaluate(self, x, y, spectral_order):

        # The spectral_order variable is coming in as an array/list of one element.
        # So, we are going to just take the 0'th element and use that as the index.
        try:
            order_number = int(spectral_order[0])
        except Exception:
            raise ValueError('Spectral order is not between 1 and 3, {}'.format(spectral_order))

        return self.models[order_number](x, y)


class Logical(Model):
    """
    Substitute values in an array where the condition is evaluated to True.

    Similar to numpy's where function.

    Parameters
    ----------
    condition : str
        A string representing the logical, one of GT, LT, NE, EQ
    compareto : float, ndarray
        A number to compare to using the condition
        If ndarray then the input array, compareto and value should have the
        same shape.
    value : float, ndarray
        Value to substitute where condition is True.
    """
    n_inputs = 1
    n_outputs = 1
    _separable = False

    conditions = {'GT': np.greater,
                  'LT': np.less,
                  'EQ': np.equal,
                  'NE': np.not_equal
                  }

    def __init__(self, condition, compareto, value, **kwargs):
        self.condition = condition.upper()
        self.compareto = compareto
        self.value = value
        super(Logical, self).__init__(**kwargs)
        self.inputs = ('x', )
        self.outputs = ('x', )

    def evaluate(self, x):
        x = x.copy()
        m = ~np.isnan(x)
        m_ind = np.flatnonzero(m)
        if isinstance(self.compareto, np.ndarray):
            cond = self.conditions[self.condition](x[m], self.compareto[m])
            x.flat[m_ind[cond]] = self.value.flat[m_ind[cond]]
        else:
            cond = self.conditions[self.condition](x[m], self.compareto)
            x.flat[m_ind[cond]] = self.value
        return x

    def __repr__(self):
        txt = "{0}(condition={1}, compareto={2}, value={3})"
        return txt.format(self.__class__.__name__, self.condition,
                          self.compareto, self.value)


class IdealToV2V3(Model):
    """
    Performs the transform from Ideal to telescope V2,V3 coordinate system.
    The two systems have the same origin: V2_REF, V3_REF.

    Note: This model has no schema implemented - add schema if needed.
    """
    _separable = False
    n_inputs = 2
    n_outputs = 2

    v3idlyangle = Parameter()  # in deg
    v2ref = Parameter()  # in arcsec
    v3ref = Parameter()  # in arcsec
    vparity = Parameter()

    def __init__(self, v3idlyangle, v2ref, v3ref, vparity, name='idl2V', **kwargs):
        super(IdealToV2V3, self).__init__(v3idlyangle=v3idlyangle, v2ref=v2ref,
                                          v3ref=v3ref, vparity=vparity, name=name,
                                          **kwargs)
        self.inputs = ('xidl', 'yidl')
        """ x and y coordinates in the telescope Ideal frame."""
        self.outputs = ('v2', 'v3')
        """ coordinates in the telescope (V2,V3) frame."""

    @staticmethod
    def evaluate(xidl, yidl, v3idlyangle, v2ref, v3ref, vparity):
        """
        Parameters
        ----------
        xidl, yidl : ndarray-like
            Coordinates in Ideal System [in arcsec]
        v3idlyangle : float
            Angle between Ideal Y-axis and V3 [ in deg]
        v2ref, v3ref : ndarray-like
            Coordinates in V2, V3 [in arcsec]
        vparity : int
            Parity.

        Returns
        -------
        v2, v3 : ndarray-like
            Coordinates in the (V2, V3) telescope system [in arcsec].

        """
        v3idlyangle = np.deg2rad(v3idlyangle)

        v2 = v2ref + vparity * xidl * np.cos(v3idlyangle) + yidl * np.sin(v3idlyangle)
        v3 = v3ref - vparity * xidl * np.sin(v3idlyangle) + yidl * np.cos(v3idlyangle)
        return v2, v3

    def inverse(self):
        return V2V3ToIdeal(self.v3idlyangle, self.v2ref, self.v3ref, self.vparity)


class V2V3ToIdeal(Model):
    """
    Performs the transform from telescope V2,V3 to Ideal coordinate system.
    The two systems have the same origin - V2_REF, V3_REF.

    Note: This model has no schema implemented - add if needed.
    """
    _separable = False

    n_inputs = 2
    n_outputs = 2

    v3idlyangle = Parameter()  # in deg
    v2ref = Parameter()  # in arcsec
    v3ref = Parameter()  # in arcsec
    vparity = Parameter()

    def __init__(self, v3idlyangle, v2ref, v3ref, vparity, name='V2idl', **kwargs):
        super(V2V3ToIdeal, self).__init__(v3idlyangle=v3idlyangle, v2ref=v2ref,
                                          v3ref=v3ref, vparity=vparity, name=name,
                                          **kwargs)
        self.inputs = ('v2', 'v3')
        """ ('v2', 'v3'): coordinates in the telescope (V2,V3) frame."""
        self.outputs = ('xidl', 'yidl')
        """ ('xidl', 'yidl'): x and y coordinates in the telescope Ideal frame."""

    @staticmethod
    def evaluate(v2, v3, v3idlyangle, v2ref, v3ref, vparity):
        """
        Parameters
        ----------
        xidl, yidl : ndarray-like
            Coordinates in Ideal System [in arcsec]
        v3idlyangle : float
            Angle between Ideal Y-axis and V3 [ in deg]
        v2ref, v3ref : ndarray-like
            Coordinates in V2, V3 [in arcsec]
        vparity : int
            Parity.

        Returns
        -------
        xidl, yidl : ndarray-like
            Coordinates in the Ideal telescope system [in arcsec].

        """
        v3idlyangle = np.deg2rad(v3idlyangle)

        xidl = vparity * ((v2 - v2ref) * np.cos(v3idlyangle) -
                          (v3 - v3ref) * np.sin(v3idlyangle))
        yidl = ((v2 - v2ref) * np.sin(v3idlyangle) +
                (v3 - v3ref) * np.cos(v3idlyangle))

        return xidl, yidl

    def inverse(self):
        return IdealToV2V3(self.v3idlyangle, self.v2ref, self.v3ref, self.vparity)


def _toindex(value):
    """
    Convert value to an int or an int array.

    Input coordinates converted to integers
    corresponding to the center of the pixel.
    The convention is that the center of the pixel is
    (0, 0), while the lower left corner is (-0.5, -0.5).

    Examples
    --------
    >>> from stdatamodels.jwst.transforms.models import _toindex
    >>> _toindex(np.array([-0.5, 0.49999]))
    array([0, 0])
    >>> _toindex(np.array([0.5, 1.49999]))
    array([1, 1])
    >>> _toindex(np.array([1.5, 2.49999]))
    array([2, 2])
    """
    indx = np.asarray(np.floor(np.asarray(value) + 0.5), dtype=int)
    return indx


class NIRCAMForwardRowGrismDispersion(Model):
    """Return the transform from grism to image for the given spectral order.

    Parameters
    ----------
    orders : list [int]
        List of orders which are available

    lmodels : list [astropy.modeling.Model]
        List of models which govern the wavelength solutions for each order

    xmodels : list [astropy.modeling.Model]
        List of models which govern the x solutions for each order

    ymodels : list [astropy.modeling.Model]
        List of models which govern the y solutions for each order

    inv_xmodels : list [astropy.modeling.Model]
        List of models which will be used if inverse ymodels
        cannot be analytically derived

    Returns
    -------
    x, y, wavelength, order in the grism image for the pixel at x0,y0 that was
    specified as input using the input delta pix for the specified order

    Notes
    -----
    The evaluation here is linear currently because higher orders have not yet been
    defined for NIRCAM (NIRCAM polynomials currently do not have any field
    dependence)
    """
    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 5
    n_outputs = 4

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, inv_lmodels=None, inv_xmodels=None,
                 inv_ymodels=None, name=None, meta=None):
        self.orders = orders
        self.lmodels = lmodels
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.inv_lmodels = inv_lmodels
        self.inv_xmodels = inv_xmodels
        self.inv_ymodels = inv_ymodels
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        meta = {"orders": orders}  # informational for users
        if name is None:
            name = 'nircam_forward_row_grism_dispersion'
        super(NIRCAMForwardRowGrismDispersion, self).__init__(name=name,
                                                              meta=meta)
        self.inputs = ("x", "y", "x0", "y0", "order")
        self.outputs = ("x", "y", "wavelength", "order")

    def evaluate(self, x, y, x0, y0, order):
        """Return the transform from grism to image for the given spectral order.

        Parameters
        ----------
        x : float
            input x pixel
        y : float
            input y pixel
        x0 : float
            input x-center of object
        y0 : float
            input y-center of object
        order : int
            the spectral order to use
        """
        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        if not self.inv_xmodels:
            t = self.invdisp_interp(iorder, x0, y0, (x - x0))
        else:
            t = self.inv_xmodels[iorder](x - x0)

        lmodel = self.lmodels[iorder]

        def apply_poly(coeff_model, inputs, t):
            # Determine order of polynomial in t
            ord_t = len(coeff_model)
            if ord_t == 1:
                if isinstance(coeff_model, (ListNode, list)):
                    sumval = coeff_model[0](t)
                else:
                    sumval = coeff_model(t)
            else:
                sumval = 0.
                for i in range(ord_t):
                    sumval += t ** i * coeff_model[i](*inputs[:coeff_model[i].n_inputs])
            return sumval

        l_poly = apply_poly(lmodel, (x0, y0), t)

        return x0, y0, l_poly, order

    def invdisp_interp(self, order, x0, y0, dx):

        if len(dx.shape) == 2:
            dx = dx[0, :]

        t_len = dx.shape[0]
        t0 = np.linspace(0., 1., t_len)

        if isinstance(self.xmodels[order], (ListNode, list)):
            if len(self.xmodels[order]) == 2:
                xr = self.xmodels[order][0](x0, y0) + t0 * self.xmodels[order][1](x0, y0)
            elif len(self.xmodels[order]) == 3:
                xr = self.xmodels[order][0](x0, y0) + t0 * self.xmodels[order][1](x0, y0) + \
                     t0**2 * self.xmodels[order][2](x0, y0)
            elif len(self.xmodels[order][0].inputs) == 1:
                xr = (dx - self.xmodels[order][0].c0.value) / self.xmodels[order][0].c1.value
                return xr
            else:
                raise Exception
        else:
            xr = (dx - self.xmodels[order].c0.value) / self.xmodels[order].c1.value
            return xr

        if len(xr.shape) > 1:
            xr = xr[0, :]

        so = np.argsort(xr)
        f = np.interp(dx, xr[so], t0[so])

        f = np.broadcast_to(f, dx.shape)
        return f


class NIRCAMForwardColumnGrismDispersion(Model):
    """Return the transform from grism to image for the given spectral order.

    Parameters
    ----------
    orders : list [int]
        List of orders which are available

    lmodels : list [astropy.modeling.Model]
        List of models which govern the wavelength solutions

    xmodels : list [astropy.modeling.Model]
        List of models which govern the x solutions

    ymodels : list [astropy.modeling.Model]
        List of models which govern the y solutions

    inv_ymodels : list [astropy.modeling.Model]
        List of models which will be used if inverse ymodels
        cannot be analytically derived

    Returns
    -------
    x, y, lam, order in the grism image for the pixel at x0,y0 that was
    specified as input using the input delta pix for the specified order

    Notes
    -----
    The evaluation here is linear because higher orders have not yet been
    defined for NIRCAM (NIRCAM polynomials currently do not have any field
    dependence)
    """
    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 5
    n_outputs = 4

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, inv_lmodels=None, inv_xmodels=None,
                 inv_ymodels=None, name=None, meta=None):
        self.orders = orders
        self.lmodels = lmodels
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.inv_lmodels = inv_lmodels
        self.inv_xmodels = inv_xmodels
        self.inv_ymodels = inv_ymodels
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        meta = {"orders": orders}  # informational for users
        if name is None:
            name = 'nircam_forward_column_grism_dispersion'
        super(NIRCAMForwardColumnGrismDispersion, self).__init__(name=name,
                                                                 meta=meta)
        self.inputs = ("x", "y", "x0", "y0", "order")
        self.outputs = ("x", "y", "wavelength", "order")

    def evaluate(self, x, y, x0, y0, order):
        """Return the transform from grism to image for the given spectral order.

        Parameters
        ----------
        x : float
            input x pixel
        y : float
            input y pixel
        x0 : float
            input x-center of object
        y0 : float
            input y-center of object
        order : int
            the spectral order to use
        """
        def apply_poly(coeff_model, inputs, t):
            # Determine order of polynomial in t
            ord_t = len(coeff_model)
            if ord_t == 1:
                if isinstance(coeff_model, (ListNode, list)):
                    sumval = coeff_model[0](t)
                else:
                    sumval = coeff_model(t)
            else:
                sumval = 0.
                for i in range(ord_t):
                    sumval += t ** i * coeff_model[i](*inputs[2-coeff_model[i].n_inputs:])
            return sumval

        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        lmodel = self.lmodels[iorder]

        if not self.inv_ymodels:
            t = self.invdisp_interp(self.ymodels, iorder, x0, y0, (y - y0))
        else:
            t = self.inv_ymodels[iorder](y - y0)

        l_poly = apply_poly(lmodel, (x0, y0), t)

        return x0, y0, l_poly, order

    def invdisp_interp(self, model, order, x0, y0, dy):

        if len(dy.shape) == 2:
            dy = dy[0, :]

        t_len = dy.shape[0]
        t0 = np.linspace(0., 1., t_len)

        if isinstance(model, (ListNode, list)):
            if len(model[order]) == 2:
                xr = model[order][0](x0, y0) + t0 * model[order][1](x0, y0)
            elif len(model[order]) == 3:
                xr = model[order][0](x0, y0) + t0 * model[order][1](x0, y0) + \
                     t0 ** 2 * model[order][2](x0, y0)
            elif len(model[order][0].inputs) == 1:
                xr = (dy - model[order][0].c0.value) / model[order][0].c1.value
                return xr
            else:
                raise Exception
        else:
            xr = (dy - model[order].c0.value) / model[order].c1.value
            return xr

        if len(xr.shape) > 1:
            xr = xr[0, :]

        so = np.argsort(xr)
        f = np.interp(dy, xr[so], t0[so])

        f = np.broadcast_to(f, dy.shape)
        return f


class NIRCAMBackwardGrismDispersion(Model):
    """Return the valid pixel(s) and wavelengths given center x,y and lam

    Parameters
    ----------
    orders : list [int]
        List of orders which are available

    lmodels : list [astropy.modeling.Model]
        List of models which govern the wavelength solutions

    xmodels : list [astropy.modeling.Model]
        List of models which govern the x solutions

    ymodels : list [astropy.modeling.Model]
        List of models which govern the y solutions

    inv_lmodels: list [astropy.modeling.Model]
        List of models which will be used if inverse lmodels
        cannot be analytically derived

    Returns
    -------
    x, y, lam, order in the grism image for the pixel at x0,y0 that was
    specified as input using the wavelength l for the specified order

    Notes
    -----
    The evaluation here is linear because higher orders have not yet been defined for NIRCAM
    (NIRCAM polynomials currently do not have any field dependence)
    """
    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 4
    n_outputs = 5

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, inv_lmodels=None, inv_xmodels=None,
                 inv_ymodels=None, name=None, meta=None):
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        self.orders = orders
        self.lmodels = lmodels
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.inv_lmodels = inv_lmodels
        self.inv_xmodels = inv_xmodels
        self.inv_ymodels = inv_ymodels
        meta = {"orders": orders}
        if name is None:
            name = "nircam_backward_grism_dispersion"
        super(NIRCAMBackwardGrismDispersion, self).__init__(name=name,
                                                            meta=meta)
        self.inputs = ("x", "y", "wavelength", "order")
        self.outputs = ("x", "y", "x0", "y0", "order")

    def evaluate(self, x, y, wavelength, order):
        """Return the transform from image to grism for the given spectral order.

        Parameters
        ----------
        x : float
            input x pixel
        y : float
            input y pixel
        wavelength : float
            input wavelength in angstroms
        order : int
            specifies the spectral order
        """
        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        if (wavelength < 0).any():
            raise ValueError("wavelength should be greater than zero")

        if not self.inv_lmodels:
            t = self.invdisp_interp(self.lmodels[iorder], x, y, wavelength)
        else:
            lmodel = self.inv_lmodels[iorder]
            t = assess_model(lmodel, x=x, y=y, t=wavelength)
        xmodel = self.xmodels[iorder]
        ymodel = self.ymodels[iorder]

        dx = assess_model(xmodel, x, y, t)
        dy = assess_model(ymodel, x, y, t)
        return x + dx, y + dy, x, y, order

    def invdisp_interp(self, model, x0, y0, wavelength):

        t0 = np.linspace(0., 1., 40)
        t_re = np.reshape(t0, [len(t0), *map(int, np.ones_like(np.shape(x0)))])

        if len(model) == 2:
            xr = (np.ones_like(t_re) * model[0](x0, y0)) + (t_re * model[1](x0, y0))
        elif len(model) == 3:
            xr = (np.ones_like(t_re) * model[0](x0, y0)) + (t_re * model[1](x0, y0)) + \
                 (t_re ** 2 * model[2](x0, y0))
        else:
            if isinstance(model, (ListNode, list)):
                xr = model[0](t0)
            else:
                xr = model(t0)
            f = np.zeros_like(wavelength)
            for i, w in enumerate(wavelength):
                f[i] = np.interp(w, xr, t0)
            return f

        so = np.argsort(xr, axis=1)
        f = np.zeros_like(wavelength)
        for i, w in enumerate(wavelength):
            f[i] = np.interp(w, np.take_along_axis(xr, so, axis=1)[:,i], t0)
        return f


class NIRISSBackwardGrismDispersion(Model):
    """This model calculates the dispersion extent of NIRISS pixels.

    The dispersion is relative to the input x,y for a given wavelength.

    Parameters
    ----------
    xmodels : list[tuple]
        The list of tuple(models) for the polynomial model in x
    ymodels : list[tuple]
        The list of tuple(models) for the polynomial model in y
    lmodels : list
        The list of models for the polynomial model in l
    orders : list
        The list of orders which are available to the model
    theta : float
        Angle [deg] - defines the NIRISS filter wheel position

    Notes
    -----
    Given the x,y, wave, order as known on the direct image,
    it returns the tuple of x, y, wave, order for that wave in the dispersed image.

    This model needs to be generalized, at the moment it satisfies the
    2t x 6(xy)th order polynomial currently used by NIRISS.

    """

    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 4
    n_outputs = 5

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, theta=0.0, name=None, meta=None):
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.lmodels = lmodels
        self.orders = orders
        self.theta = theta
        meta = {"orders": orders}
        if name is None:
            name = 'niriss_backward_grism_dispersion'
        super(NIRISSBackwardGrismDispersion, self).__init__(name=name,
                                                            meta=meta)
        self.inputs = ("x", "y", "wavelength", "order")
        self.outputs = ("x", "y", "x0", "y0", "order")

    def evaluate(self, x, y, wavelength, order):
        """Return the valid pixel(s) and wavelengths given center x,y and lam

        Parameters
        ----------
        wavelength : float
            Input wavelength you want to know about, will be converted to float
        x : float
            Input x location
        y : float
            Input y location
        wavelength : float
            Wavelength to disperse
        order : list
            The order to use


        Returns
        -------
        x, y, wavelength, order in the grism image for the pixel at x,y that was
        specified as input using the wavelength and order specified

        Notes
        -----
        There's spatial dependence for NIRISS so the forward transform
        depends on x,y as well as the filter wheel rotation. Theta is
        usu. taken to be the different between fwcpos_ref in the specwcs
        reference file and fwcpos from the input image.

        """
        if (wavelength < 0).any():
            raise ValueError("Wavelength should be greater than zero")
        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        t = self.lmodels[iorder](wavelength)
        xmodel = self.xmodels[iorder]
        ymodel = self.ymodels[iorder]

        dx = xmodel[0](x, y) + t * xmodel[1](x, y) + t**2 * xmodel[2](x, y)
        dy = ymodel[0](x, y) + t * ymodel[1](x, y) + t**2 * ymodel[2](x, y)

        # rotate by theta
        if self.theta != 0.0:
            rotate = Rotation2D(self.theta)
            dx, dy = rotate(dx, dy)

        return x + dx, y + dy, x, y, order


class NIRISSForwardRowGrismDispersion(Model):
    """This model calculates the wavelengths of vertically dispersed NIRISS grism data.

    The dispersion polynomial is relative to the input x,y pixels
    in the direct image for a given wavelength.

    Parameters
    ----------
    orders : list
        The list of orders which are available to the model
    xmodels : list[tuples]
        The list of tuple(models) for the polynomial model in x
    ymodels : list[tuples]
        The list of tuple(models) for the polynomial model in y
    lmodels : list
        The list of models for the polynomial model in l
    theta : float
        Angle [deg] - defines the NIRISS filter wheel position

    Notes
    -----
    Given the x,y, source location as known on the dispersed image, as well as order,
    it returns the tuple of x,y,wavelength,order.

    This model needs to be generalized, at the moment it satisfies the
    2t x 6(xy)th order polynomial currently used by NIRISS.

    """

    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 5
    n_outputs = 4

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, theta=0., name=None, meta=None):
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.lmodels = lmodels
        self.theta = theta
        self.orders = orders
        meta = {"orders": orders}
        if name is None:
            name = 'niriss_forward_row_grism_dispersion'
        super(NIRISSForwardRowGrismDispersion, self).__init__(name=name,
                                                              meta=meta)
        # starts with the backwards pixel and calculates the forward pixel
        self.inputs = ("x", "y", "x0", "y0", "order")
        self.outputs = ("x", "y", "wavelength", "order")

    def evaluate(self, x, y, x0, y0, order):
        """Return the valid pixel(s) and wavelengths given center x,y and lam

        Parameters
        ----------
        x0: int,float,list
            Source object x-center

        y0: int,float,list
            Source object y-center

        x :  int,float,list
            Input x location

        y :  int,float,list
            Input y location

        order : int
            Spectral order to use


        Returns
        -------
        x, y, lambda, order, theta,  in the direct image for the pixel that was
        specified as input using the wavelength l and spectral order

        Notes
        -----
        There's spatial dependence for NIRISS as well as dependence on the
        filter wheel rotation during the exposure.

        """
        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        # The next two lines are to get around the fact that
        # modeling.standard_broadcasting=False does not work.
        x00 = x0.flatten()[0]
        y00 = y0.flatten()[0]

        t = np.linspace(0, 1, 10)  # sample t
        xmodel = self.xmodels[iorder]
        ymodel = self.ymodels[iorder]
        lmodel = self.lmodels[iorder]

        dx = xmodel[0](x00, y00) + t * xmodel[1](x00, y00) + t**2 * xmodel[2](x00, y00)
        dy = ymodel[0](x00, y00) + t * ymodel[1](x00, y00) + t**2 * ymodel[2](x00, y00)

        if self.theta != 0.0:
            rotate = Rotation2D(self.theta)
            dx, dy = rotate(dx, dy)

        so = np.argsort(dx)
        tab = Tabular1D(dx[so], t[so], bounds_error=False, fill_value=None)

        dxr = astmath.SubtractUfunc()
        wavelength = dxr | tab | lmodel
        model = Mapping((2, 3, 0, 2, 4)) | Const1D(x00) & Const1D(y00) & wavelength & Const1D(order)
        # returns x0, y0, lam, order
        return model(x, y, x0, y0, order)


class NIRISSForwardColumnGrismDispersion(Model):
    """This model calculates the wavelengths for horizontally dispersed NIRISS grism data.

    The dispersion polynomial is relative to the input x,y pixels
    in the direct image for a given wavelength.

    Parameters
    ----------
    orders : list
        The list of orders which are available to the model.
    xmodels : list[tuple]
        The list of tuple(models) for the polynomial model in x
    ymodels : list[tuple]
        The list of tuple(models) for the polynomial model in y
    lmodels : list
        The list of models for the polynomial model in wavelength.
    theta : float
        Angle [deg] - defines the NIRISS filter wheel position

    Notes
    -----
    Given the x,y, source location, order, it returns the tuple of
    x,y,wavelength,order on the dispersed image. It also requires
    FWCPOS from the image header, this is the filter wheel position
    in degrees.

    """

    standard_broadcasting = False
    _separable = False
    fittable = False
    linear = False

    n_inputs = 5
    n_outputs = 4

    def __init__(self, orders, lmodels=None, xmodels=None,
                 ymodels=None, theta=None, name=None, meta=None):
        self._order_mapping = {int(k): v for v, k in enumerate(orders)}
        self.xmodels = xmodels
        self.ymodels = ymodels
        self.lmodels = lmodels
        self.orders = orders
        self.theta = theta
        meta = {"orders": orders}
        if name is None:
            name = 'niriss_forward_column_grism_dispersion'
        super(NIRISSForwardColumnGrismDispersion, self).__init__(name=name,
                                                                 meta=meta)
        # starts with the backwards pixel and calculates the forward pixel
        self.inputs = ("x", "y", "x0", "y0", "order")
        self.outputs = ("x", "y", "wavelength", "order")

    def evaluate(self, x, y, x0, y0, order):
        """Return the valid pixel(s) and wavelengths given center x,y and lam

        Parameters
        ----------
        x0: int,float
            Source object x-center
        y0: int,float
            Source object y-center
        x :  int,float
            Input x location
        y :  int,float
            Input y location
        order : int
            Spectral order to use
        theta : float
            input filter wheel rotation angle in degrees

        Returns
        -------
        x, y, lambda, order,  in the direct image for the pixel that was
        specified as input using the wavelength l and spectral order

        Notes
        -----
        There's spatial dependence for NIRISS as well as rotation for the filter wheel

        """
        try:
            iorder = self._order_mapping[int(order.flatten()[0])]
        except KeyError:
            raise ValueError("Specified order is not available")

        # The next two lines are to get around the fact that
        # modeling.standard_broadcasting=False does not work.
        x00 = x0.flatten()[0]
        y00 = y0.flatten()[0]

        t = np.linspace(0, 1, 10)
        xmodel = self.xmodels[iorder]
        ymodel = self.ymodels[iorder]
        lmodel = self.lmodels[iorder]
        dx = xmodel[0](x00, y00) + t * xmodel[1](x00, y00) + t**2 * xmodel[2](x00, y00)
        dy = ymodel[0](x00, y00) + t * ymodel[1](x00, y00) + t**2 * ymodel[2](x00, y00)

        if self.theta != 0.0:
            rotate = Rotation2D(self.theta)
            dx, dy = rotate(dx, dy)
        so = np.argsort(dy)
        tab = Tabular1D(dy[so], t[so], bounds_error=False, fill_value=None)
        dyr = astmath.SubtractUfunc()
        wavelength = dyr | tab | lmodel
        model = Mapping((2, 3, 1, 3, 4)) | Const1D(x00) & Const1D(y00) & wavelength & Const1D(order)
        return model(x, y, x0, y0, order)


class Rotation3DToGWA(Model):
    """
    Perform a 3D rotation given an angle in degrees.
    Positive angles represent a counter-clockwise rotation and vice-versa.
    Parameters
    ----------
    angles : array-like
        Angles of rotation in deg in the order of axes_order.
    axes_order : str
        A sequence of 'x', 'y', 'z' corresponding of axis of rotation/
    """
    standard_broadcasting = False
    _separable = False

    separable = False

    n_inputs = 3
    n_outputs = 3

    angles = Parameter(getter=np.rad2deg, setter=np.deg2rad)

    def __init__(self, angles, axes_order, name=None):
        if len(angles) != len(axes_order):
            raise InputParameterError(
                "Number of angles must equal number of axes in axes_order.")

        self.axes = ['x', 'y', 'z']
        unrecognized = set(axes_order).difference(self.axes)
        if unrecognized:
            raise ValueError("Unrecognized axis label {0}; "
                             "should be one of {1} ".format(unrecognized,
                                                            self.axes))
        self.axes_order = axes_order

        self._func_map = {'x': self._xrot,
                          'y': self._yrot,
                          'z': self._zrot
                          }
        super(Rotation3DToGWA, self).__init__(angles, name=name)
        self.inputs = ('x', 'y', 'z')
        self.outputs = ('x', 'y', 'z')

    @property
    def inverse(self):
        """Inverse rotation."""
        angles = self.angles.value[::-1] * -1
        return self.__class__(angles, self.axes_order[::-1])

    def _xrot(self, x, y, z, theta):
        xout = x
        yout = y * np.cos(theta) + z * np.sin(theta)
        zout = np.sqrt(1 - xout ** 2 - yout ** 2)
        return [xout, yout, zout]

    def _yrot(self, x, y, z, theta):
        xout = x * np.cos(theta) - z * np.sin(theta)
        yout = y
        zout = np.sqrt(1 - xout ** 2 - yout ** 2)
        return [xout, yout, zout]

    def _zrot(self, x, y, z, theta):
        xout = x * np.cos(theta) + y * np.sin(theta)
        yout = -x * np.sin(theta) + y * np.cos(theta)
        zout = np.sqrt(1 - xout ** 2 - yout ** 2)
        return [xout, yout, zout]

    def evaluate(self, x, y, z, angles):
        """
        Apply the rotation to a set of 3D Cartesian coordinates.
        """

        if x.shape != y.shape != z.shape:
            raise ValueError("Expected input arrays to have the same shape")

        #  Note: If the original shape was () (an array scalar) convert to a
        #  1-element 1-D array on output for consistency with most other models
        orig_shape = x.shape or (1,)
        for ang, ax in zip(angles[0], self.axes_order):
            x, y, z = self._func_map[ax](x, y, z, theta=ang)
        x.shape = y.shape = z.shape = orig_shape

        return x, y, z


class Snell(Model):
    """
    Apply transforms, including Snell law, through the NIRSpec prism.
    Parameters
    ----------
    angle : float
        Prism angle in deg.
    kcoef : list
        K coefficients in Sellmeier equation.
    lcoef : list
        L coefficients in Sellmeier equation.
    tcoef : list
        Thermal coefficients of glass.
    tref : float
        Reference temperature in K.
    pref : float
        Reference pressure in ATM.
    temperature : float
        System temperature during observation in K
    pressure : float
        System pressure during observation in ATM.
    """

    standard_broadcasting = False
    _separable = False

    n_inputs = 4
    n_outputs = 3

    def __init__(self, angle, kcoef, lcoef, tcoef, tref, pref,
                 temperature, pressure, name=None):
        self.prism_angle = angle
        self.kcoef = np.array(kcoef, dtype=float)
        self.lcoef = np.array(lcoef, dtype=float)
        self.tcoef = np.array(tcoef, dtype=float)
        self.tref = tref
        self.pref = pref
        self.temp = temperature
        self.pressure = pref
        super(Snell, self).__init__(name=name)
        self.inputs = ("lam", "alpha_in", "beta_in", "zin")
        self.outputs = ("alpha_out", "beta_out", "zout")

    @staticmethod
    def compute_refraction_index(lam, temp, tref, pref, pressure, kcoef, lcoef, tcoef):
        """Calculate and return the refraction index."""

        # Convert to microns
        lam = np.asarray(lam * 1e6)
        KtoC = 273.15  # kelvin to celsius conversion
        temp -= KtoC
        tref -= KtoC
        delt = temp - tref

        K1, K2, K3 = kcoef
        L1, L2, L3 = lcoef
        D0, D1, D2, E0, E1, lam_tk = tcoef

        if delt < 20:
            n = np.sqrt(1. +
                        K1 * lam**2 / (lam**2 - L1) +
                        K2 * lam**2 / (lam**2 - L2) +
                        K3 * lam**2 / (lam**2 - L3)
                        )
        else:
            # Derive the refractive index of air at the reference temperature and pressure
            # and at the operational system's temperature and pressure.
            nref = 1. + (6432.8 + 2949810. * lam**2 /
                         (146.0 * lam**2 - 1.) + (5540.0 * lam**2) /
                         (41.0 * lam**2 - 1.)) * 1e-8

            # T should be in C, P should be in ATM
            nair_obs = 1.0 + ((nref - 1.0) * pressure) / (1.0 + (temp - 15.) * 3.4785e-3)
            nair_ref = 1.0 + ((nref - 1.0) * pref) / (1.0 + (tref - 15) * 3.4785e-3)

            # Compute the relative index of the glass at Tref and Pref using Sellmeier equation I.
            lamrel = lam * nair_obs / nair_ref

            nrel = np.sqrt(1. +
                           K1 * lamrel ** 2 / (lamrel ** 2 - L1) +
                           K2 * lamrel ** 2 / (lamrel ** 2 - L2) +
                           K3 * lamrel ** 2 / (lamrel ** 2 - L3)
                           )
            # Convert the relative index of refraction at the reference temperature and pressure
            # to absolute.
            nabs_ref = nrel * nair_ref

            # Compute the absolute index of the glass
            delnabs = (0.5 * (nrel ** 2 - 1.) / nrel) * \
                (D0 * delt + D1 * delt ** 2 + D2 * delt ** 3 +
                 (E0 * delt + E1 * delt ** 2) / (lamrel ** 2 - lam_tk ** 2))
            nabs_obs = nabs_ref + delnabs

            # Define the relative index at the system's operating T and P.
            n = nabs_obs / nair_obs
        return n

    def evaluate(self, lam, alpha_in, beta_in, zin):
        """Go through the prism"""
        n = self.compute_refraction_index(lam, self.temp, self.tref, self.pref, self.pressure,
                                          self.kcoef, self.lcoef, self.tcoef)
        # Apply Snell's law through front surface, eq 5.3.3 II
        xout = alpha_in / n
        yout = beta_in / n
        zout = np.sqrt(1.0 - xout**2 - yout**2)

        # Go to back surface frame # eq 5.3.3 III
        y_rotation = Rotation3DToGWA([self.prism_angle], "y")
        xout, yout, zout = y_rotation(xout, yout, zout)

        # Reflection on back surface
        xout = -1 * xout
        yout = -1 * yout

        # Back to front surface
        y_rotation = Rotation3DToGWA([-self.prism_angle], "y")
        xout, yout, zout = y_rotation(xout, yout, zout)

        # Snell's refraction law through front surface
        xout = xout * n
        yout = yout * n
        zout = np.sqrt(1.0 - xout**2 - yout**2)
        return xout, yout, zout


class AngleFromGratingEquation(Model):
    """
    Solve the 3D Grating Dispersion Law for the refracted angle.

    Parameters
    ----------
    groove_density : int
        Grating ruling density.
    order : int
        Spectral order.
    """

    _separable = False
    n_inputs = 4
    n_outputs = 3

    groove_density = Parameter()
    """ Grating ruling density."""

    order = Parameter(default=-1)
    """ Spectral order."""

    def __init__(self, groove_density, order, **kwargs):
        super().__init__(groove_density=groove_density, order=order, **kwargs)
        self.inputs = ("lam", "alpha_in", "beta_in", "z")
        """ Wavelength and 3 angle coordinates going into the grating."""

        self.outputs = ("alpha_out", "beta_out", "zout")
        """ Three angles coming out of the grating. """

    def evaluate(self, lam, alpha_in, beta_in, z, groove_density, order):
        if alpha_in.shape != beta_in.shape != z.shape:
            raise ValueError("Expected input arrays to have the same shape")
        orig_shape = alpha_in.shape or (1,)
        xout = -alpha_in - groove_density * order * lam
        yout = - beta_in
        zout = np.sqrt(1 - xout**2 - yout**2)
        xout.shape = yout.shape = zout.shape = orig_shape
        return xout, yout, zout


class WavelengthFromGratingEquation(Model):
    """
    Solve the 3D Grating Dispersion Law for the wavelength.

    Parameters
    ----------
    groove_density : int
        Grating ruling density.
    order : int
        Spectral order.
    """

    _separable = False
    n_inputs = 3
    n_outputs = 1

    groove_density = Parameter()
    """ Grating ruling density."""
    order = Parameter(default=1)
    """ Spectral order."""

    def __init__(self, groove_density, order, **kwargs):
        super().__init__(groove_density=groove_density, order=order, **kwargs)
        self.inputs = ("alpha_in", "beta_in", "alpha_out")
        """ three angle - alpha_in and beta_in going into the grating and alpha_out coming out of the grating."""
        self.outputs = ("lam",)
        """ Wavelength."""

    def evaluate(self, alpha_in, beta_in, alpha_out, groove_density, order):
        # beta_in is not used in this equation but is here because it's
        # needed for the prism computation. Currently these two computations
        # need to have the same interface.
        return -(alpha_in + alpha_out) / (groove_density * order)


class Rotation3D(Model):
    """
    Perform a series of rotations about different axis in 3D space.

    Positive angles represent a counter-clockwise rotation.

    Parameters
    ----------
    angles : array-like
        Angles of rotation in deg in the order of axes_order.
    axes_order : str
        A sequence of 'x', 'y', 'z' corresponding of axis of rotation.
    """
    standard_broadcasting = False
    _separable = False

    n_inputs = 3
    n_outputs = 3

    angles = Parameter(getter=np.rad2deg, setter=np.deg2rad)

    def __init__(self, angles, axes_order, name=None):
        self.axes = ['x', 'y', 'z']
        unrecognized = set(axes_order).difference(self.axes)
        if unrecognized:
            raise ValueError("Unrecognized axis label {0}; "
                             "should be one of {1} ".format(unrecognized,
                                                            self.axes))
        self.axes_order = axes_order
        if len(angles) != len(axes_order):
            raise ValueError("The number of angles {0} should match the number \
                              of axes {1}.".format(len(angles),
                                                   len(axes_order)))
        super(Rotation3D, self).__init__(angles, name=name)
        self.inputs = ('x', 'y', 'z')
        self.outputs = ('x', 'y', 'z')

    @property
    def inverse(self):
        """Inverse rotation."""
        angles = self.angles.value[::-1] * -1
        return self.__class__(angles, axes_order=self.axes_order[::-1])

    @staticmethod
    def _compute_matrix(angles, axes_order):
        if len(angles) != len(axes_order):
            raise InputParameterError(
                "Number of angles must equal number of axes in axes_order.")
        matrices = []
        for angle, axis in zip(angles, axes_order):
            matrix = np.zeros((3, 3), dtype=float)
            if axis == 'x':
                mat = Rotation3D.rotation_matrix_from_angle(angle)
                matrix[0, 0] = 1
                matrix[1:, 1:] = mat
            elif axis == 'y':
                mat = Rotation3D.rotation_matrix_from_angle(-angle)
                matrix[1, 1] = 1
                matrix[0, 0] = mat[0, 0]
                matrix[0, 2] = mat[0, 1]
                matrix[2, 0] = mat[1, 0]
                matrix[2, 2] = mat[1, 1]
            elif axis == 'z':
                mat = Rotation3D.rotation_matrix_from_angle(angle)
                matrix[2, 2] = 1
                matrix[:2, :2] = mat
            else:
                raise ValueError(f"""Expected axes_order to be a combination
                                 of characters 'x', 'y' and 'z',
                                 got {set(axes_order).difference(['x', 'y', 'z'])}""")
            matrices.append(matrix)
        if len(angles) == 1:
            return matrix
        elif len(matrices) == 2:
            return np.dot(matrices[1], matrices[0])
        else:
            prod = np.dot(matrices[1], matrices[0])
            for m in matrices[2:]:
                prod = np.dot(m, prod)
            return prod

    @staticmethod
    def rotation_matrix_from_angle(angle):
        """
        Clockwise rotation matrix.
        """
        return np.array([[math.cos(angle), -math.sin(angle)],
                         [math.sin(angle), math.cos(angle)]])

    def evaluate(self, x, y, z, angles):
        """
        Apply the rotation to a set of 3D Cartesian coordinates.
        """
        if x.shape != y.shape != z.shape:
            raise ValueError("Expected input arrays to have the same shape")
        # Note: If the original shape was () (an array scalar) convert to a
        # 1-element 1-D array on output for consistency with most other models
        orig_shape = x.shape or (1,)
        inarr = np.array([x.flatten(), y.flatten(), z.flatten()])
        result = np.dot(self._compute_matrix(angles[0], self.axes_order), inarr)
        x, y, z = result[0], result[1], result[2]
        x.shape = y.shape = z.shape = orig_shape
        return x, y, z


class V23ToSky(Rotation3D):
    """
    Transform from V2V3 to a standard coordinate system (ICRS).

    Parameters
    ----------
    angles : list
        A sequence of angles (in deg).
        The angles are [-V2_REF, V3_REF, -ROLL_REF, -DEC_REF, RA_REF].
    axes_order : str
        A sequence of characters ('x', 'y', or 'z') corresponding to the
        axis of rotation and matching the order in ``angles``.
        The axes are "zyxyz".
    """

    _separable = False

    n_inputs = 2
    n_outputs = 2

    def __init__(self, angles, axes_order, name=None):
        super(V23ToSky, self).__init__(angles, axes_order=axes_order, name=name)
        self._inputs = ("v2", "v3")
        """ ("v2", "v3"): Coordinates in the (V2, V3) telescope frame."""
        self._outputs = ("ra", "dec")
        """ ("ra", "dec"): RA, DEC coordinates in ICRS."""

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, val):
        self._inputs = val

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, val):
        self._outputs = val

    @staticmethod
    def spherical2cartesian(alpha, delta):
        """
        Convert spherical coordinates (in deg) to cartesian.
        """
        alpha = np.deg2rad(alpha)
        delta = np.deg2rad(delta)
        x = np.cos(alpha) * np.cos(delta)
        y = np.cos(delta) * np.sin(alpha)
        z = np.sin(delta)
        return np.array([x, y, z])

    @staticmethod
    def cartesian2spherical(x, y, z):
        """
        Convert cartesian coordinates to spherical coordinates (in deg).
        """
        h = np.hypot(x, y)
        alpha = np.rad2deg(np.arctan2(y, x))
        delta = np.rad2deg(np.arctan2(z, h))
        return alpha, delta

    def evaluate(self, v2, v3, angles):
        x, y, z = self.spherical2cartesian(v2, v3)
        x1, y1, z1 = super(V23ToSky, self).evaluate(x, y, z, angles)
        ra, dec = self.cartesian2spherical(x1, y1, z1)

        return ra, dec

    def __call__(self, v2, v3, **kwargs):
        from itertools import chain
        inputs, format_info = self.prepare_inputs(v2, v3)
        parameters = self._param_sets(raw=True)

        outputs = self.evaluate(*chain(inputs, parameters))
        if self.n_outputs == 1:
            outputs = (outputs,)

        return self.prepare_outputs(format_info, *outputs)


class Unitless2DirCos(Model):
    """
    Transform a vector to directional cosines.
    """
    _separable = False
    n_inputs = 2
    n_outputs = 3

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.inputs = ('x', 'y')
        self.outputs = ('x', 'y', 'z')

    def evaluate(self, x, y):
        vabs = np.sqrt(1. + x**2 + y**2)
        cosa = x / vabs
        cosb = y / vabs
        cosc = 1. / vabs
        return cosa, cosb, cosc

    def inverse(self):
        return DirCos2Unitless()


class DirCos2Unitless(Model):
    """
    Transform directional cosines to vector.
    """
    _separable = False
    n_inputs = 3
    n_outputs = 2

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.inputs = ('x', 'y', 'z')
        self.outputs = ('x', 'y')

    def evaluate(self, x, y, z):

        return x / z, y / z

    def inverse(self):
        return Unitless2DirCos()

def assess_model(model, x=0, y=0, t=0):
    if isinstance(model, (ListNode, list)):
        ninputs = len(model[0].inputs)
        if ninputs == 2:
            output = model[0](x, y) + t * model[1](x, y) + t ** 2 * model[2](x, y)
        elif ninputs == 1:
            if len(model) == 1:
                output = model[0](t)
            elif len(model) == 2:
                output = model[0](x) + t * model[1](x)
        else:
            raise ValueError(f"{model} has incorrect number of inputs required.")
    else:
        ninputs = len(model.inputs)
        if ninputs == 2:
            output = model(x, y)
        elif ninputs == 1:
            output = model(t)
        else:
            raise ValueError(f"{model} has incorrect number of inputs required.")
    return output



from .reference import ReferenceFileModel

__all__ = ['OutlierParsModel']


class OutlierParsModel(ReferenceFileModel):
    """
    A data model for outlier detection parameters reference tables.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/outlierpars.schema"



"""
Various utility functions and data types
"""

import os

import numpy as np
from astropy.io import fits
import asdf
from asdf import treeutil

try:
    from asdf.treeutil import RemoveNode
except ImportError:
    # Prior to asdf 2.8, None was used to indicate
    # that a node should be removed.
    RemoveNode = None

import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def gentle_asarray(a, dtype):
    """
    Performs an asarray that doesn't cause a copy if the byteorder is
    different.  It also ignores column name differences -- the
    resulting array will have the column names from the given dtype.
    """
    out_dtype = np.dtype(dtype)
    if isinstance(a, np.ndarray):
        in_dtype = a.dtype
        # Non-table array
        if in_dtype.fields is None and out_dtype.fields is None:
            if np.can_cast(in_dtype, out_dtype, 'equiv'):
                return a
            else:
                return _safe_asanyarray(a, out_dtype)
        elif in_dtype.fields is not None and out_dtype.fields is not None:
            # When a FITS file includes a pseudo-unsigned-int column, astropy will return
            # a FITS_rec with an incorrect table dtype.  The following code rebuilds
            # in_dtype from the individual fields, which are correctly labeled with an
            # unsigned int dtype.
            # We can remove this once the issue is resolved in astropy:
            # https://github.com/astropy/astropy/issues/8862
            if isinstance(a, fits.fitsrec.FITS_rec):
                a.dtype = rebuild_fits_rec_dtype(a)
                in_dtype = a.dtype

            if in_dtype == out_dtype:
                return a
            in_names = {n.lower() for n in in_dtype.names}
            out_names = {n.lower() for n in out_dtype.names}
            if in_names == out_names:
                # Change the dtype name to match the fits record names
                # as the mismatch causes case insensitive access to fail
                out_dtype.names = in_dtype.names
            else:
                raise ValueError(
                    "Column names don't match schema. "
                    "Schema has {0}. Data has {1}".format(
                        str(out_names.difference(in_names)),
                        str(in_names.difference(out_names))))

            new_dtype = []
            for i in range(len(out_dtype.fields)):
                in_type = in_dtype[i]
                out_type = out_dtype[i]
                if in_type.subdtype is None:
                    type_str = in_type.str
                else:
                    type_str = in_type.subdtype[0].str
                if np.can_cast(in_type, out_type, 'equiv'):
                    new_dtype.append(
                        (out_dtype.names[i],
                         type_str,
                         in_type.shape))
                else:
                    return _safe_asanyarray(a, out_dtype)
            return a.view(dtype=np.dtype(new_dtype))
        else:
            return _safe_asanyarray(a, out_dtype)
    else:
        try:
            a = np.asarray(a, dtype=out_dtype)
        except Exception:
            raise ValueError("Can't convert {0!s} to ndarray".format(type(a)))
        return a


def _safe_asanyarray(a, dtype):
    if isinstance(a, fits.fitsrec.FITS_rec):
        if any(c.bzero is not None for c in a.columns):
            # Due to an issue in astropy, it's not safe to directly cast
            # a FITS_rec with a pseudo-unsigned column.
            # See https://github.com/astropy/astropy/issues/12112
            result = np.zeros(a.shape, dtype=dtype)
            for old_col, new_col in zip(a.dtype.names, result.dtype.names):
                result[new_col] = a[old_col]
            return result

    return np.asanyarray(a, dtype=dtype)


def create_history_entry(description, software=None):
    """
    Create a HistoryEntry object.

    Parameters
    ----------
    description : str
        Description of the change.
    software : dict or list of dict
        A description of the software used.  It should not include
        asdf itself, as that is automatically notated in the
        `asdf_library` entry.

        Each dict must have the following keys:

        ``name``: The name of the software
        ``author``: The author or institution that produced the software
        ``homepage``: A URI to the homepage of the software
        ``version``: The version of the software

    Examples
    --------
    >>> soft = {'name': 'jwreftools', 'author': 'STSCI', \
                'homepage': 'https://github.com/spacetelescope/jwreftools', 'version': "0.7"}
    >>> entry = create_history_entry(description="HISTORY of this file", software=soft)

    """
    from asdf.tags.core import Software, HistoryEntry
    import datetime

    if isinstance(software, list):
        software = [Software(x) for x in software]
    elif software is not None:
        software = Software(software)

    entry = HistoryEntry({
        'description': description,
        'time': datetime.datetime.utcnow()
    })

    if software is not None:
        entry['software'] = software
    return entry


def get_envar_as_boolean(name, default=False):
    """Interpret an environmental as a boolean flag

    Truth is any numeric value that is not 0 or
    any of the following case-insensitive strings:

    ('true', 't', 'yes', 'y')

    Parameters
    ----------
    name : str
        The name of the environmental variable to retrieve

    default : bool
        If the environmental variable cannot be accessed, use as the default.
    """
    truths = ('true', 't', 'yes', 'y')
    falses = ('false', 'f', 'no', 'n')
    if name in os.environ:
        value = os.environ[name]
        try:
            value = bool(int(value))
        except ValueError:
            value_lowcase = value.lower()
            if value_lowcase not in truths + falses:
                raise ValueError(f'Cannot convert value "{value}" to boolean unambiguously.')
            return value_lowcase in truths
        return value

    log.debug(f'Environmental "{name}" cannot be found. Using default value of "{default}".')
    return default


def get_model_type(init):
    """
    Fetch the model type string from the underlying file object.

    Parameters
    ----------
    init : asdf.AsdfFile or astropy.io.fits.HDUList

    Returns
    -------
    str or None
    """
    if isinstance(init, asdf.AsdfFile):
        if "meta" in init:
            return init["meta"].get("model_type")
        else:
            return None
    elif isinstance(init, fits.HDUList):
        return init[0].header.get("DATAMODL")
    else:
        raise TypeError(f"Unhandled init type: {init.__class__.__name__}")


def remove_none_from_tree(tree):
    """
    Remove None values from a tree.  Both dictionary keys
    and list indices with None values will be removed.

    Parameters
    ----------
    tree : object
        The root node of the tree.

    Returns
    -------
    object
        Modified tree.
    """

    def _remove_none(node):
        if node is None:
            return RemoveNode
        else:
            return node

    return treeutil.walk_and_modify(tree, _remove_none)


def rebuild_fits_rec_dtype(fits_rec):
    dtype = fits_rec.dtype
    new_dtype = []
    for field_name in dtype.fields:
        table_dtype = dtype[field_name]
        field_dtype = fits_rec.field(field_name).dtype
        if np.issubdtype(table_dtype, np.signedinteger) and np.issubdtype(field_dtype, np.unsignedinteger):
            new_dtype.append((field_name, field_dtype))
        else:
            new_dtype.append((field_name, table_dtype))
    return np.dtype((np.record, new_dtype))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import numpy as np
import os
import warnings
from astropy.modeling.models import Shift, Rotation2D, Const1D

from astropy.utils import minversion

if minversion("asdf_astropy", "0.3.0", False):
    from asdf_astropy.testing.helpers import assert_model_roundtrip
else:
    from asdf_astropy.converters.transform.tests.test_transform import assert_model_roundtrip

from stdatamodels.jwst.transforms.models import (
    NirissSOSSModel, Rotation3DToGWA, Gwa2Slit, Logical, Slit,
    DirCos2Unitless, Unitless2DirCos, Snell, AngleFromGratingEquation,
    WavelengthFromGratingEquation)
import asdf
import pytest


m1 = Shift(1) & Shift(2) | Rotation2D(3.1)
m2 = Shift(2) & Shift(2) | Rotation2D(23.1)


test_models = [DirCos2Unitless(), Unitless2DirCos(),
               Rotation3DToGWA(angles=[12.1, 1.3, 0.5, 3.4], axes_order='xyzx'),
               AngleFromGratingEquation(20000, -1), WavelengthFromGratingEquation(25000, 2),
               Logical('GT', 5, 10), Logical('LT', np.ones((10,)) * 5, np.arange(10)),
               Snell(angle=-16.5, kcoef=[0.583, 0.462, 3.891], lcoef=[0.002526, 0.01, 1200.556],
                     tcoef=[-2.66e-05, 0.0, 0.0, 0.0, 0.0, 0.0], tref=35, pref=0,
                     temperature=35, pressure=0),
               ]


@pytest.mark.parametrize(('model'), test_models)
def test_model(tmpdir, model, version=None):
    assert_model_roundtrip(model, tmpdir)


def test_gwa_to_slit(tmpdir):
    transforms = [m1, m2]
    s0 = Slit("s0", 1, 2, 3, 4, 5, 6, 7, 8)
    s1 = Slit("s1", 10, 20, 30, 40, 50, 60, 70, 80)
    slits = [s0, s1]
    m = Gwa2Slit(slits, transforms)
    assert_model_roundtrip(m, tmpdir)

    slits = [1, 2]
    m = Gwa2Slit(slits, transforms)
    assert_model_roundtrip(m, tmpdir)


def test_niriss_soss(tmpdir):
    """Regression test for bugs discussed in issue #7401"""
    spectral_orders = [1, 2, 3]
    models = [
        Const1D(1.0) & Const1D(2.0) & Const1D(3.0),
        Const1D(4.0) & Const1D(5.0) & Const1D(6.0),
        Const1D(7.0) & Const1D(8.0) & Const1D(9.0),
    ]

    soss_model = NirissSOSSModel(spectral_orders, models)

    # Check that no warning is issued when serializing
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert_model_roundtrip(soss_model, tmpdir)

    # Check tag is the latest version
    path = str(tmpdir / "test.asdf")
    with asdf.AsdfFile({"model": soss_model}) as af:
        af.write_to(path)

    with asdf.open(path, _force_raw_types=True) as af:
        assert af.tree["model"]._tag == "tag:stsci.edu:jwst_pipeline/niriss_soss-1.0.0"


def test_niriss_soss_legacy():
    data_path = os.path.join(os.path.dirname(__file__), 'data')
    data = os.path.join(data_path, 'niriss_soss.asdf')

    # confirm that the file contains the legacy tag
    with asdf.open(data, _force_raw_types=True) as af:
        assert af.tree["model"]._tag == "tag:stsci.edu:jwst_pipeline/niriss-soss-0.7.0"

    # test that it opens with the legacy tag
    with asdf.open(data) as af:
        model = af['model']
        assert model.spectral_orders == [1, 2, 3]
        assert (model.models[1].parameters == (1.0, 2.0, 3.0)).all()
        assert (model.models[2].parameters == (4.0, 5.0, 6.0)).all()
        assert (model.models[3].parameters == (7.0, 8.0, 9.0)).all()



from .reference import ReferenceFileModel

__all__ = ['Extract1dIFUModel']


class Extract1dIFUModel(ReferenceFileModel):
    """
    A data model for IFU MIRI and NIRSpec extract 1d reference files.

    Parameters
    __________
    extract1d_params : numpy table
        Basic extract 1D parameters
        - region_type: ascii
        - subtract_background: bool
        - method: ascii
        - subpixels: int16
    extract1d_table : numpy table
        extract1d parameters for varying wavelengths
        A table-like object containing extract 1d parameters
        based on wavelength
        - wavelength: float32 1D array
        - radius: float32 1D array
        - inner_bkg: float32 1D array
        - outer_bkg: float32 1D array
        - axis_ratio: float32 1D array
        - axis_pa: float32 1D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/extract1difu.schema"



import numpy as np
from stdatamodels import util


def test_gentle_asarray():
    x = np.array([('abc', 1.0)], dtype=[
        ('FOO', 'S3'),
        ('BAR', '>f8')])

    new_dtype = [('foo', '|S3'), ('bar', '<f8')]

    y = util.gentle_asarray(x, new_dtype)

    assert y['BAR'][0] == 1.0



"""Project default for pytest"""
from pathlib import Path

import pytest

import asdf


def pytest_addoption(parser):
    parser.addoption(
        "--no-crds",
        action="store_true",
        default=False,
        help="Skip tests against crds",
    )

@pytest.fixture(scope="session", autouse=True)
def register_schemas():
    schemas_root = Path(__file__).parent/"schemas"
    with asdf.config_context() as config:
        config.add_resource_mapping(
            asdf.resource.DirectoryResourceMapping(schemas_root, "http://example.com/schemas")
        )
        yield


@pytest.fixture(autouse=True)
def patch_env_variables(monkeypatch):
    """
    Make sure the environment doesn't initially contain these so
    that test results are consistent.
    """
    for var in ["PASS_INVALID_VALUES", "STRICT_VALIDATION", "SKIP_FITS_UPDATE", "VALIDATE_ON_ASSIGNMENT"]:
        monkeypatch.delenv(var, raising=False)



import warnings

from .image import ImageModel

__all__ = ['DrizProductModel']


def DrizProductModel(*args, **kwargs):
    warnings.simplefilter('default')
    warnings.warn(message="DrizProduct is deprecated and will be removed.  "
                  "Use ImageModel.", category=DeprecationWarning)
    return ImageModel(*args, **kwargs)



import warnings

from .multislit import MultiSlitModel


__all__ = ['MultiProductModel']


def MultiProductModel(*args, **kwargs):
    warnings.simplefilter('default')
    warnings.warn(message="MultiProductModel is deprecated and will be removed.  "
                  "Use MultiSlitModel.", category=DeprecationWarning)
    return MultiSlitModel(*args, **kwargs)



from astropy.io import fits
from astropy.time import Time
import jsonschema
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from stdatamodels.jwst.datamodels import MultiSlitModel, ImageModel


def test_multislit_model():
    array1 = np.asarray(np.random.rand(2, 2), dtype='float32')
    array2 = np.asarray(np.random.rand(2, 2), dtype='float32')

    with MultiSlitModel() as ms:
        assert len(ms.slits) == 0
        ms.slits.append(ms.slits.item())
        ms.slits[-1].data = array1
        assert len(ms.slits) == 1
        ms.slits.append(ms.slits.item())
        ms.slits[-1].data = array2
        assert len(ms.slits) == 2
        for i, slit in enumerate(ms.slits):
            assert slit == ms.slits[i]
        ms2 = ms.copy()
        assert len(ms2.slits) == 2
        assert_array_equal(ms.slits[-1].data, array2)
        del ms.slits[0]
        assert len(ms.slits) == 1
        assert_array_equal(ms.slits[0].data, array2)


def test_multislit_move_from_fits(tmp_path):
    path = tmp_path / "multislit_move.fits"

    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    for i in range(5):
        hdu = fits.ImageHDU(data=np.zeros((64, 64)), name='SCI')
        hdu.ver = i + 1
        hdulist.append(hdu)

    hdulist.writeto(path)

    n = MultiSlitModel()
    with MultiSlitModel(path) as m:
        n.slits.append(m.slits[2])

        assert len(n.slits) == 1


def test_multislit_append_string():
    with pytest.raises(jsonschema.ValidationError):
        m = MultiSlitModel(strict_validation=True)
        m.slits.append('junk')


def test_multislit():
    with MultiSlitModel() as dm:
        dm.slits.append(dm.slits.item())
        slit = dm.slits[-1]
        slit.data = np.random.rand(5, 5)
        slit.dm = np.random.rand(5, 5)
        slit.err = np.random.rand(5, 5)
        assert slit.wavelength.shape == (0, 0)
        assert slit.pathloss_point.shape == (0, 0)
        assert slit.pathloss_uniform.shape == (0, 0)
        assert slit.barshadow.shape == (0, 0)


def test_multislit_from_image():
    with ImageModel((64, 64)) as im:
        with MultiSlitModel(im) as ms:
            assert len(ms.slits) == 1
            assert ms.slits[0].data.shape == (64, 64)


def test_multislit_from_saved_imagemodel(tmp_path):
    path = tmp_path / "multislit_from_image.fits"
    with ImageModel((64, 64)) as im:
        im.save(path)

    with MultiSlitModel(path) as ms:
        assert len(ms.slits) == 1
        assert ms.slits[0].data.shape == (64, 64)

        for i, slit in enumerate(ms.slits):
            assert slit.data is ms.slits[i].data

        ms2 = ms.copy()
        ms2.slits = ms.slits
        assert len(ms2.slits) == 1


def test_multislit_metadata():
    with MultiSlitModel() as ms:
        with ImageModel((64, 64)) as im:
            ms.slits.append(ms.slits.item())
            ms.slits[-1].data = im.data
        slit = ms.slits[0]
        slit.name = "FOO"
        assert ms.slits[0].name == "FOO"


def test_multislit_metadata2():
    with MultiSlitModel() as ms:
        ms.slits.append(ms.slits.item())
        for key, val in ms.items():
            assert isinstance(val, (bytes, str, int, float, bool, Time))


def test_multislit_copy(tmp_path):
    path = tmp_path / "multislit.fits"
    with MultiSlitModel() as input:
        for i in range(4):
            input.slits.append(input.slits.item(data=np.empty((50, 50))))

        assert len(input.slits) == 4
        input.save(path)

        output = input.copy()
        assert len(output.slits) == 4

    with fits.open(path, memmap=False) as hdulist:
        assert len(hdulist) == 6

    with MultiSlitModel(path) as model:
        for i, slit in enumerate(model.slits):
            pass
        assert i + 1 == 4

        output = model.copy()
        assert len(output.slits) == 4


def test_copy_multislit():
    model1 = MultiSlitModel()
    model2 = MultiSlitModel()

    model1.slits.append(ImageModel(np.ones((1024, 1024))))
    model2.slits.append(ImageModel(np.ones((1024, 1024)) * 2))

    # Create the output model as a copy of the first input
    output = model1.copy()

    assert len(model1.slits) == 1
    assert len(model2.slits) == 1
    assert len(output.slits) == 1

    assert model1.slits[0].data[330, 330] == 1
    assert output.slits[0].data[330, 330] == 1
    assert id(model1.slits[0].data) != id(output.slits[0].data)

    output.slits[0].data = model1.slits[0].data - model2.slits[0].data

    assert model1.slits[0].data[330, 330] == 1
    assert output.slits[0].data[330, 330] == -1



from .model_base import JwstDataModel


__all__ = ['OutlierIFUOutputModel']


class OutlierIFUOutputModel(JwstDataModel):
    """
    A data model for the optional output from outlier_detection_ifu  step.

    In the parameter definitions below, `n` is the number of
    exposures,  `ny` and `nx` are the height and width of the image.

    Parameters
    __________

    diffarr : numpy float32 array (n, ny, nx)
        Minimum difference array for all the data

    minarr : numpy float32 array (ny, nx)
        Final combined minimum difference array

    normarr : numpy float32 array (ny, nx)
        Normalized minarr

    minnorm : numpy float32 array (ny, nx)
        minarr/normarr

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/outlierifuoutput.schema"



from .model_base import JwstDataModel
from .image import ImageModel


__all__ = ['SlitModel', 'SlitDataModel']


class SlitDataModel(JwstDataModel):
    """
    A data model for 2D slit images.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    var_poisson : numpy float32 array
         variance due to poisson noise

    var_rnoise : numpy float32 array
         variance due to read noise

    var_flat : numpy float32 array
         variance due to flat

    wavelength : numpy float32 array
         Wavelength array, corrected for zero-point

    barshadow : numpy float32 array
         Bar shadow correction

    flatfield_point : numpy float32 array
         flatfield array for point source

    flatfield_uniform : numpy float32 array
         flatfield array for uniform source

    pathloss_point : numpy float32 array
         pathloss array for point source

    pathloss_uniform : numpy float32 array
         pathloss array for uniform source

    photom_point : numpy float32 array
         photom array for point source

    photom_uniform : numpy float32 array
         photom array for uniform source

    area : numpy float32 array
         Pixel area map array
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/slitdata.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, (SlitModel, ImageModel)):
            super(SlitDataModel, self).__init__(init=None, **kwargs)
            self.data = init.data
            self.dq = init.dq
            self.err = init.err
            self.area = init.area
            if init.hasattr('var_poisson'):
                self.var_poisson = init.var_poisson
            if init.hasattr('var_rnoise'):
                self.var_rnoise = init.var_rnoise
            if init.hasattr('var_flat'):
                self.var_flat = init.var_flat
            if init.hasattr('wavelength'):
                self.wavelength = init.wavelength
            for key in kwargs:
                setattr(self, key, kwargs[key])

            if init.meta.hasattr('wcs'):
                self.meta.wcs = init.meta.wcs
            else:
                self.meta.wcs = None
        else:
            super(SlitDataModel, self).__init__(init=init, **kwargs)
            if kwargs:
                for key in kwargs:
                    setattr(self, key, kwargs[key])


class SlitModel(JwstDataModel):
    """
    A data model for 2D images.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    var_poisson : numpy float32 array
         variance due to poisson noise

    var_rnoise : numpy float32 array
         variance due to read noise

    var_flat : numpy float32 array
         variance due to flat

    wavelength : numpy float32 array
         Wavelength array, corrected for zero-point

    barshadow : numpy float32 array
         Bar shadow correction

    flatfield_point : numpy float32 array
         flatfield array for point source

    flatfield_uniform : numpy float32 array
         flatfield array for uniform source

    pathloss_point : numpy float32 array
         pathloss array for point source

    pathloss_uniform : numpy float32 array
         pathloss array for uniform source

    photom_point : numpy float32 array
         photom array for point source

    photom_uniform : numpy float32 array
         photom array for uniform source

    area : numpy float32 array
         Pixel area map array

    int_times : numpy table
         table of times for each integration
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/slit.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, (SlitModel, ImageModel)):
            super(SlitModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.data = init.data
            self.dq = init.dq
            self.err = init.err
            self.area = init.area
            if init.hasattr('var_poisson'):
                self.var_poisson = init.var_poisson
            if init.hasattr('var_rnoise'):
                self.var_rnoise = init.var_rnoise
            if init.hasattr('var_flat'):
                self.var_flat = init.var_flat
            if init.hasattr('wavelength'):
                self.wavelength = init.wavelength
            if init.hasattr('int_times'):
                self.int_times = init.int_times
            if init.meta.hasattr('wcs'):
                self.meta.wcs = init.meta.wcs
            else:
                self.meta.wcs = None
        else:
            super(SlitModel, self).__init__(init=init, **kwargs)
            if kwargs:
                for key in kwargs:
                    setattr(self, key, kwargs[key])



from .model_base import JwstDataModel


__all__ = ['ImageModel']


class ImageModel(JwstDataModel):
    """
    A data model for 2D images.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    zeroframe : numpy float32 array
         Zeroframe array

    var_poisson : numpy float32 array
         variance due to poisson noise

    var_rnoise : numpy float32 array
         variance due to read noise

    area : numpy float32 array
         Pixel area map array

    pathloss_point : numpy float32 array
         Pathloss correction for point source

    pathloss_uniform : numpy float32 array
         Pathloss correction for uniform source
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/image.schema"



"""General utility objects"""


def deprecate_class(new_class,
                    message='"{old_class}" is deprecated and will be removed. Use {new_class}'):
    """Deprecate a class in favor of another class

    Parameters
    ----------
    new_class : class
        The class/object that should be used instead

    message : str
        The deprecation warning message
    """

    # Implement the inner decorator
    def _decorator(old_class):

        def init(self, *args, **kwargs):
            import warnings
            warnings.simplefilter('default')
            warnings.warn(message.format(old_class=old_class.__name__, new_class=new_class.__name__),
                          category=DeprecationWarning)
            new_class.__init__(self, *args, **kwargs)

        # Create the class
        deprecated = type(
            old_class.__name__,
            (new_class,),
            {'__doc__': old_class.__doc__,
             '__init__': init,
             '__module__': old_class.__module__,
             '__name__': old_class.__name__}
        )

        return deprecated

    return _decorator


def bytes2human(n):
    """Convert bytes to human-readable format

    Taken from the `psutil` library which references
    http://code.activestate.com/recipes/578019

    Parameters
    ----------
    n : int
        Number to convert

    Returns
    -------
    readable : str
        A string with units attached.

    Examples
    --------
    >>> bytes2human(10000)
        '9.8K'

    >>> bytes2human(100001221)
        '95.4M'
    """
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return '%.1f%s' % (value, s)
    return "%sB" % n



from .model_base import JwstDataModel


__all__ = ['IFUCubeModel']


class IFUCubeModel(JwstDataModel):
    """
    A data model for 3D IFU  cubes.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    weightmap : numpy float32 array
         Weight map of coverage

    wavetable : numpy table
         Wavelength value for slices
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ifucube.schema"

    def __init__(self, init=None, **kwargs):
        super(IFUCubeModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Test jwst.transforms
"""
import pytest
from numpy.testing import assert_allclose

from stdatamodels.jwst.transforms import models


# _RANDOM_SEED = 0x1337

"""
def test_logical():

    with NumpyRNGContext(0x1337):
        compareto = np.random.randn(10)
    with NumpyRNGContext(0x1338):
        val = np.random.randn(10)
    with NumpyRNGContext(0x1339):
        x = np.random.randn(10)
    l = models.Logical('GT', .5, 10)

    res = l(x)
    y = x.copy()
    y[np.greater(x, .5)] = 10
    assert_allclose(res, npres)
    l = models.Logical('lt', compareto, val)
    cond = np.less(x, compareto)
    y = x.copy()
    y[cond] = val[cond]
    assert_allclose(res, npres)
"""


def test_ideal_to_v23_roundtrip():
    """
    Test roundtripping of the transforms.
    """
    v2i = models.V2V3ToIdeal(.4, 450, 730, 1)
    x, y = 450, 730
    assert_allclose(v2i.inverse(*v2i(x, y)), (x, y))


@pytest.mark.parametrize(('wavelength', 'n'),
                         [(1e-6, 1.43079543),
                          (2e-6, 1.42575377),
                          (5e-6, 1.40061966)
                          ])
def test_refraction_index(wavelength, n):
    """
    Tests the computation of the refraction index.
    True values are from the ESA pipeline.
    Reference values are from the PRISM reference file from CV3.
    """
    temp_sys = 37.06107795068881  # in K
    tref = 35  # in K
    pref = 0  # in atm
    pressure_sys = 0  # in atm
    kcoef = [0.58339748, 0.46085267, 3.8915394]
    lcoef = [0.00252643, 0.010078333, 1200.556]
    tcoef = [-2.66e-05, 0.0, 0.0, 0.0, 0.0, 0.0]
    n_pipeline = models.Snell.compute_refraction_index(wavelength, temp_sys,
                                                       tref, pref, pressure_sys,
                                                       kcoef, lcoef, tcoef)
    assert_allclose(n_pipeline, n)



"""
Functions that support validation of model changes
"""

import warnings
import jsonschema
from asdf import schema as asdf_schema
from asdf import yamlutil
from asdf.tags.core import ndarray
from asdf.schema import ValidationError, YAML_VALIDATORS
from asdf.util import HashableDict
import numpy as np

from .util import remove_none_from_tree


class ValidationWarning(Warning):
    pass


def value_change(path, value, schema, ctx):
    """
    Validate a change in value against a schema.
    Trap error and return a flag.
    """
    try:
        _check_value(value, schema, ctx)
        update = True

    except jsonschema.ValidationError as error:
        update = False
        errmsg = _error_message(path, error)
        if ctx._pass_invalid_values:
            update = True
        if ctx._strict_validation:
            raise jsonschema.ValidationError(errmsg)
        else:
            warnings.warn(errmsg, ValidationWarning)
    return update


def _validate_datatype(validator, schema_datatype, instance, schema):
    """
    This extends the ASDF datatype validator to support ndim
    and max_ndim within individual fields of structured arrays,
    and handle the absence of the shape field correctly.

    Additionally, dtypes are required to be equivalent, instead
    of just "safe" to cast.
    """
    if isinstance(instance, list):
        array = ndarray.inline_data_asarray(instance)
        instance_datatype, _ = ndarray.numpy_dtype_to_asdf_datatype(array.dtype)
    elif isinstance(instance, dict):
        if 'datatype' in instance:
            instance_datatype = instance['datatype']
        elif 'data' in instance:
            array = ndarray.inline_data_asarray(instance['data'])
            instance_datatype, _ = ndarray.numpy_dtype_to_asdf_datatype(array.dtype)
        else:
            yield ValidationError("Not an array")
    elif isinstance(instance, (np.ndarray, ndarray.NDArrayType)):
        instance_datatype, _ = ndarray.numpy_dtype_to_asdf_datatype(instance.dtype)
    else:
        yield ValidationError("Not an array")

    schema_dtype = ndarray.asdf_datatype_to_numpy_dtype(schema_datatype)
    instance_dtype = ndarray.asdf_datatype_to_numpy_dtype(instance_datatype)

    if not schema_dtype.fields:
        if instance_dtype.fields:
            yield ValidationError(f"Expected scalar datatype '{schema_datatype}', got '{instance_datatype}'")

        # Using 'equiv' so that we can be flexible on byte order:
        if not np.can_cast(instance_dtype, schema_dtype, 'equiv'):
            yield ValidationError(f"Array datatype '{instance_datatype}' is not compatible with '{schema_datatype}'")
    else:
        if not instance_dtype.fields:
            yield ValidationError(f"Expected structured datatype '{schema_datatype}', got '{instance_datatype}'")

        if len(instance_dtype.fields) != len(schema_dtype.fields):
            yield ValidationError(
                "Mismatch in number of fields: "
                f"Expected {len(schema_datatype)}, got {len(instance_datatype)}"
            )

        for i in range(len(schema_dtype.fields)):
            instance_type = instance_dtype[i]
            instance_ndim = len(instance_type.shape)
            schema_type = schema_dtype[i]
            field_schema = schema_datatype[i]
            instance_name = instance_dtype.names[i]
            schema_name = schema_dtype.names[i]

            if instance_name != schema_name:
                yield ValidationError(
                    f"Wrong name in field {i}: Expected "
                    f"{instance_name}, got {schema_name}"
                )

            if 'ndim' in field_schema and instance_ndim != field_schema['ndim']:
                yield ValidationError(
                    f"Wrong number of dimensions in field {i}: Expected "
                    f"{field_schema['ndim']}, got {instance_ndim}"
                )

            if 'max_ndim' in field_schema and instance_ndim > field_schema['max_ndim']:
                yield ValidationError(
                    f"Wrong number of dimensions in field {i}: Expected "
                    f"maximum of {field_schema['max_ndim']}, got {instance_ndim}"
                )

            if len(schema_type.shape) == 0:
                # If the schema didn't include a shape, then anything goes:
                compare_instance_type = instance_type.base
                compare_schema_type = schema_type.base
            else:
                # If shape is present in the schema, then the array must match:
                compare_instance_type = instance_type
                compare_schema_type = schema_type

            # Using 'equiv' so that we can be flexible on byte order:
            if not np.can_cast(compare_instance_type, compare_schema_type, 'equiv'):
                yield ValidationError(
                    f"Array datatype '{instance_datatype}' is not compatible with '{schema_datatype}' at "
                    f"field {i}"
                )


_VALIDATORS = HashableDict(YAML_VALIDATORS.copy())
_VALIDATORS["datatype"] = _validate_datatype
_VALIDATORS["ndim"] = ndarray.validate_ndim
_VALIDATORS["max_ndim"] = ndarray.validate_max_ndim


def _check_value(value, schema, ctx):
    """
    Perform the actual validation.
    """
    # Do not validate None values.  These are regarded as missing in DataModel,
    # and will eventually be stripped out when the model is saved to FITS or ASDF.
    if value is not None:
        # There may also be Nones hiding within the value.  Do this before
        # converting to tagged tree, so that we don't have to descend unnecessarily
        # into nodes for custom types.
        value = remove_none_from_tree(value)
        value = yamlutil.custom_tree_to_tagged_tree(value, ctx._asdf)

        if ctx._validate_arrays:
            validators = _VALIDATORS
        else:
            validators = YAML_VALIDATORS

        asdf_schema.validate(value, schema=schema, validators=validators)


def _error_message(path, error):
    """
    Add the path to the attribute as context for a validation error
    """
    if isinstance(path, list):
        spath = [str(p) for p in path]
        name = '.'.join(spath)
    else:
        name = str(path)

    error = str(error)
    if len(error) > 2000:
        error = error[0:1996] + " ..."
    errfmt = "While validating {} the following error occurred:\n{}"
    errmsg = errfmt.format(name, error)
    return errmsg



import os

from .model_base import JwstDataModel


__all__ = ['AsnModel']


class AsnModel(JwstDataModel):
    """
    A data model for association tables.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/asn.schema"
    supported_formats = ['yaml', 'json', 'fits']

    def __init__(self, init=None, **kwargs):
        super(AsnModel, self).__init__(init=init, **kwargs)

        # apply logic to identify output product
        self.parse_table()

    def parse_table(self):
        self.output = None
        self.output_rootname = None
        self.inputs = []
        self.input_rootnames = []
        self.num_inputs = 0

        if not len(self.asn_table):
            return

        for i, etype in enumerate(self.asn_table.exptype):
            if 'prod' in etype.lower():
                self.output_rootname = self.asn_table.expname[i]
                for fmt in self.supported_formats:
                    fname = "{0}.{1}".format(self.output_rootname, fmt)
                    if os.path.exists(fname):
                        self.output = fname
                        break
            else:
                rootname = self.asn_table.expname[i]
                for fmt in self.supported_formats:
                    fname = "{0}.{1}".format(rootname, fmt)
                    if os.path.exists(fname):
                        self.inputs.append(fname)
                        break
                self.input_rootnames.append(rootname)
                self.num_inputs += 1



import warnings

from stdatamodels.validate import ValidationWarning
from .reference import ReferenceFileModel

__all__ = ['ABVegaOffsetModel']


class ABVegaOffsetModel(ReferenceFileModel):
    """
    A data model containing offsets to convert from AB to Vega
    magnitudes.

    Parameters
    ----------
    abvega_offset : `astropy.table.Table`
        An astropy table containing offsets to convert from AB to Vega
        magnitudes.  The ``abvega_offset`` column represents m_AB -
        m_Vega.

        There are three types of tables, depending on the instrument, each
        with different column selectors.  The columns names and data types
        are:

        * FGS

          - detector: str
          - abvega_offset: float32

        * NIRCam and NIRISS

          - filter: str
          - pupil: str
          - abvega_offset: float32

        * MIRI

          - filter: str
          - abvega_offset: float32
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/abvegaoffset.schema"

    def validate(self):
        super(ABVegaOffsetModel, self).validate()
        try:
            assert len(self.abvega_offset) > 0
            assert (self.meta.instrument.name in
                    ('FGS', 'MIRI', 'NIRCAM', 'NIRISS'))
            assert 'abvega_offset' in self.abvega_offset.colnames

            if self.meta.instrument.name == 'FGS':
                assert 'detector' in self.abvega_offset.colnames

            if self.meta.instrument.name == 'MIRI':
                assert 'filter' in self.abvega_offset.colnames

            if self.meta.instrument.name in ('NIRCAM', 'NIRISS'):
                assert 'filter' in self.abvega_offset.colnames
                assert 'pupil' in self.abvega_offset.colnames

        except AssertionError as errmsg:
            if self._strict_validation:
                raise AssertionError(errmsg)
            else:
                warnings.warn(str(errmsg), ValidationWarning)
